import threading
import inspect
import ctypes
import time
import os

from contrabass.core.FacadeUtils import FacadeUtils

TASK_READ_MODEL = "READ_MODEL"
TASK_SAVE_DEM = "SAVE_DEM"
TASK_SAVE_FVA = "SAVE_FVA"
TASK_SAVE_FVA_DEM = "SAVE_FVA_DEM"
TASK_SPREADSHEET = "SPREADSHEET"
TASK_SENSIBILITY = "TASK_SENSIBILITY"
TASK_SAVE_SPREADSHEET = "SAVE_SPREADSHEET"
TASK_FIND_AND_REMOVE_DEM = "TASK_FIND_AND_REMOVE_DEM"
TASK_SAVE_MODEL = "TASK_SAVE_MODEL"
TASK_FVA = "TASK_FVA"


class ThreadInterrupt(Exception):
    def __str__(self):
        return "Exception: Thread stopped"

    def _async_raise(tid, exctype, syserr=True):
        """Raises an exception in the threads with id tid"""
        if not inspect.isclass(exctype):
            raise TypeError("Only types can be raised (not instances)")
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(tid), ctypes.py_object(exctype)
        )
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # "if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect"
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
        if syserr:
            raise SystemError("PyThreadState_SetAsyncExc failed")


class FacadeThread(threading.Thread):
    def get_my_tid(self):
        """determines this (self's) thread id

        CAREFUL : this function is executed in the context of the caller
        thread, to get the identity of the thread represented by this
        instance.
        """
        if not self.isAlive():
            raise threading.ThreadError("the thread is not active")

        # do we have it cached?
        if hasattr(self, "_thread_id"):
            return self._thread_id

        # no, look for it in the _active dict
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid

        raise AssertionError("could not determine the thread's id")

    def raiseExc(self, exctype, tid=None):
        """Raises the given exception type in the context of this thread.

        If the thread is busy in a system call (time.sleep(),
        socket.accept(), ...), the exception is simply ignored.

        If you are sure that your exception should terminate the thread,
        one way to ensure that it works is:

                t = ThreadWithExc( ... )
                ...
                t.raiseExc( SomeException )
                while t.isAlive():
                        time.sleep( 0.1 )
                        t.raiseExc( SomeException )

        If the exception is to be caught by the thread, you need a way to
        check that your thread has caught it.

        CAREFUL : this function is executed in the context of the
        caller thread, to raise an excpetion in the context of the
        thread represented by this instance.
        """
        if tid is None:
            _async_raise(self.get_my_tid(), exctype)
        else:
            _async_raise(tid, exctype)

    def __init__(self, model_path):
        self.model_path = model_path
        super(FacadeThread, self).__init__()

    def set_task(
        self,
        task,
        notify_function=None,
        args1=None,
        args2=None,
        output_path=None,
        objective=None,
        fraction=1.0,
    ):
        self.task = task
        self.notify_function = notify_function
        self.args1 = args1
        self.args2 = args2
        self.output_path = output_path
        self.objective = objective
        self.fraction = fraction

    def run(self):
        try:
            if self.task == TASK_READ_MODEL:
                f = FacadeUtils()
                (
                    result,
                    error,
                    model,
                    model_id,
                    reactions,
                    metabolites,
                    genes,
                    reactions_list,
                ) = f.read_model(self.model_path)
                self.notify_function(
                    result,
                    error,
                    model_id,
                    reactions,
                    metabolites,
                    genes,
                    reactions_list,
                    self.args1,
                    self.args2,
                )

            elif self.task == TASK_FIND_AND_REMOVE_DEM:
                f = FacadeUtils()
                self.model = f.find_and_remove_dem(self.model_path)
                if self.output_path is not None and self.model is not None:
                    f.save_model(self.output_path, self.model)
                self.notify_function(
                    "text", self.args1, self.args2, ended=True, result=True, error=None
                )

            elif self.task == TASK_SAVE_FVA:
                f = FacadeUtils()
                (self.model, error) = f.run_fva(
                    self.model_path, objective=self.objective, fraction=self.fraction
                )
                result_ok = error == ""
                if result_ok:
                    if self.output_path is not None and self.model is not None:
                        f.save_model(self.output_path, self.model)
                    self.notify_function(
                        "text",
                        self.args1,
                        self.args2,
                        ended=True,
                        result=True,
                        error=None,
                    )
                else:
                    self.notify_function(
                        "text",
                        self.args1,
                        self.args2,
                        ended=True,
                        result=False,
                        error=error,
                    )

            elif self.task == TASK_SAVE_FVA_DEM:
                f = FacadeUtils()
                (self.model, error) = f.run_fva_remove_dem(
                    self.model_path, objective=self.objective, fraction=self.fraction
                )
                result_ok = error == ""
                if result_ok:
                    if self.output_path is not None and self.model is not None:
                        f.save_model(self.output_path, self.model)
                    self.notify_function(
                        "text",
                        self.args1,
                        self.args2,
                        ended=True,
                        result=True,
                        error=None,
                    )
                else:
                    self.notify_function(
                        "text",
                        self.args1,
                        self.args2,
                        ended=True,
                        result=False,
                        error=error,
                    )

            elif self.task == TASK_SPREADSHEET:
                f = FacadeUtils()
                s = f.run_summary_model(
                    self.model_path,
                    self.notify_function,
                    self.args1,
                    None,
                    objective=self.objective,
                    fraction=self.fraction,
                )
                self.spreadsheet = s
                if self.output_path is not None:
                    (result_ok, error) = f.save_spreadsheet(self.output_path, s)
                self.notify_function(
                    "text", self.args1, self.args2, ended=True, result=True, error=None
                )

            elif self.task == TASK_SENSIBILITY:
                f = FacadeUtils()
                s = f.run_sensibility_analysis(
                    self.model_path,
                    self.notify_function,
                    self.args1,
                    None,
                    objective=self.objective,
                )
                self.spreadsheet = s
                if self.output_path is not None:
                    (result_ok, error) = f.save_spreadsheet(self.output_path, s)
                self.notify_function(
                    "text", self.args1, self.args2, ended=True, result=True, error=None
                )

            elif self.task == TASK_SAVE_SPREADSHEET:
                f = FacadeUtils()
                s = self.spreadsheet
                if self.output_path is not None and s is not None:
                    (result_ok, error) = f.save_spreadsheet(self.output_path, s)
                    self.notify_function(
                        "text",
                        self.args1,
                        self.args2,
                        ended=True,
                        result=result_ok,
                        error=error,
                    )

            elif self.task == TASK_SAVE_MODEL:
                f = FacadeUtils()
                (result, text) = f.save_model(self.output_path, self.model)
                self.notify_function(
                    "text",
                    self.args1,
                    self.args2,
                    ended=True,
                    result=result,
                    error=text,
                )
        except Exception as error:
            # Thread stopped
            # This raise is just for debugging purposes
            # print("DEBUG: please remove in FacadeThread.py:", str(error))
            # raise error
            # print("Error:", str(error))
            pass



class CriticalCPConfig:
    def __init__(self):
        self.__model_path = None
        self.__print_f = None
        self.__args1 = None
        self.__args2 = None
        self.__output_path_spreadsheet = None
        self.__output_path_html = None
        self.__objective = None
        self.__fraction = None
        self.__processes = None

    @property
    def model_path(self):
        return self.__model_path

    @model_path.setter
    def model_path(self, value):
        self.__model_path = value

    @property
    def print_f(self):
        return self.__print_f

    @print_f.setter
    def print_f(self, value):
        self.__print_f = value

    @property
    def args1(self):
        return self.__args1

    @args1.setter
    def args1(self, value):
        self.__args1 = value

    @property
    def args2(self):
        return self.__args2

    @args2.setter
    def args2(self, value):
        self.__args2 = value

    @property
    def output_path_spreadsheet(self):
        return self.__output_path_spreadsheet

    @output_path_spreadsheet.setter
    def output_path_spreadsheet(self, value):
        self.__output_path_spreadsheet = value

    @property
    def output_path_html(self):
        return self.__output_path_html

    @output_path_html.setter
    def output_path_html(self, value):
        self.__output_path_html = value

    @property
    def objective(self):
        return self.__objective

    @objective.setter
    def objective(self, value):
        self.__objective = value

    @property
    def fraction(self):
        return self.__fraction

    @fraction.setter
    def fraction(self, value):
        self.__fraction = value

    @property
    def processes(self):
        return self.__processes

    @processes.setter
    def processes(self, value):
        self.__processes = value



MSG_SUCCESFULLY_SAVE = "File successfully saved at: {}"
MSG_UNKNOWN_ERROR = "Error, something went wrong: {}"
MSG_FURTHER_INFO = "run contrabass -h for further information."
MSG_ERROR_DECIMAL = (
    "Error: parameter 'fraction' must be a decimal number between 0.0 and 1.0"
)


def __check_model_file(path):
    if len(path) <= 4:  # Error: file has no name
        return False

    if path[-4:] != ".xml" and path[-5:] != ".json" and path[-4:] != ".yml":
        return False
    else:
        return True


def __check_spread_file(path):
    if len(path) <= 4:  # Error: file has no name
        return False

    if path[-4:] != ".xls" and path[-4:] != ".ods" and path[-5:] != ".xlsx":
        return False
    else:
        return True


def __check_html_file(path):
    if len(path) <= 5:  # Error: file has no name
        return False

    if path[-5:] != ".html":
        return False
    else:
        return True


def validate_input_model(input_file):
    if not __check_model_file(input_file):
        print("Model file must be either .xml .json .yml")
        print(MSG_FURTHER_INFO)
        exit(1)


# Method for validating the output file path for operation that generate a new model
def validate_output_model(output_path):
    if not __check_model_file(output_path):
        print("Output model file must be either .xml .json .yml")
        print(MSG_FURTHER_INFO)
        exit(1)


def validate_number(parameter):
    try:
        parameter = float(parameter)
        if parameter < 0.0 or parameter > 1.0:
            print(MSG_ERROR_DECIMAL)
            exit(1)
    except:
        print(MSG_ERROR_DECIMAL)
        exit(1)
    return parameter


# Method for validating the output file path for operation that generate a new spreadsheet file
def validate_output_spread(output_path):
    if not __check_spread_file(output_path):
        print("Output file must be .xls .xlsx or .ods")
        print(MSG_FURTHER_INFO)
        exit(1)


def validate_output_html_path(output_path):
    if not __check_html_file(output_path):
        print("Output html report file format must be .html")
        print(MSG_FURTHER_INFO)
        exit(1)


def save_final_spreadsheet(facade, output_path, verbose_f, error):
    if error == "":
        (result_ok, text) = facade.save_spreadsheet(False, output_path, verbose_f)
        if result_ok:
            print(MSG_SUCCESFULLY_SAVE.format(text))
        else:
            print(MSG_UNKNOWN_ERROR.format(text))
    else:
        print(MSG_UNKNOWN_ERROR.format(error))


def save_final_model(facade, output_path, verbose_f):
    (result_ok, text) = facade.save_model(output_path, False, verbose_f)
    if result_ok:
        print(MSG_SUCCESFULLY_SAVE.format(text))
    else:
        print(MSG_UNKNOWN_ERROR.format(text))


def check_error_save_final_model(error, facade, output_path, verbose_f):
    if error != "":
        print(MSG_UNKNOWN_ERROR.format(error))
    else:
        save_final_model(facade, output_path, verbose_f)


def verbose_f(text, args1=False, args2=False):
    if args1:
        print(text)



import os
import warnings
import os
from pathlib import Path

from contrabass.core.FacadeUtils import FacadeUtils
from contrabass.core.FacadeThread import FacadeThread, ThreadInterrupt
from contrabass.core.utils.GrowthDependentCPConfig import *
from contrabass.core.utils.CustomLogger import *
from contrabass.core.util import *

import threading
import inspect
import ctypes
import time

TASK_READ_MODEL = "READ_MODEL"
TASK_SAVE_DEM = "SAVE_DEM"
TASK_SAVE_FVA = "SAVE_FVA"
TASK_SAVE_FVA_DEM = "SAVE_FVA_DEM"
TASK_SPREADSHEET = "SPREADSHEET"
TASK_SENSIBILITY = "TASK_SENSIBILITY"
TASK_SAVE_SPREADSHEET = "SAVE_SPREADSHEET"
TASK_FIND_AND_REMOVE_DEM = "TASK_FIND_AND_REMOVE_DEM"
TASK_SAVE_MODEL = "TASK_SAVE_MODEL"
TASK_FVA = "TASK_FVA"


class Facade:
    model_path = None
    model = None
    spreadsheet = None

    model_id = None
    reactions = None
    metabolites = None
    genes = None

    thread1 = None
    tid = None

    __processes = None

    @property
    def processes(self):
        return self.__processes

    @processes.setter
    def processes(self, processes):
        self.__processes = processes

    def isAlive_tid(self, tid_input):
        for tid, tobj in threading._active.items():
            if tid == tid_input:
                return True
        return False

    def stop_tid(self, tid):
        try:
            ThreadInterrupt._async_raise(tid, ThreadInterrupt, syserr=False)
            print("STOPPING")
            while self.isAlive_tid(tid):
                time.sleep(0.1)
                print("STOPPING")
                ThreadInterrupt._async_raise(tid, ThreadInterrupt, syserr=False)
        except Exception as errr:
            # Thread stopped - catch some derivated errors
            # print("DEBUG: PLEASE REMOVE THIS RAISE IN Facade.py")
            # raise errr
            pass

    def get_tid(self):
        return self.thread1.get_my_tid()

    def stop(self):
        try:
            tid = self.get_tid()
            self.stop_tid(tid)
        except Exception as exc:
            # Thread is already stopped
            # print("DEBUG: PLEASE REMOVE THIS RAISE IN Facade.py")
            # raise exc
            pass

    def set_model_path(self, model_path):
        self.model_path = model_path

    def read_model(
        self, stoppable, model_path, notify_function, args1=None, args2=None
    ):
        self.model_path = model_path
        if not stoppable:
            (
                result,
                error,
                model,
                model_id,
                reactions,
                metabolites,
                genes,
            ) = FacadeUtils.read_model(model_path)
            self.model = model
            self.model_id = model_id
            self.reactions = reactions
            self.metabolites = metabolites
            self.genes = genes
            return (result, error)
        else:
            self.thread1 = FacadeThread(self.model_path)
            self.thread1.set_task(TASK_READ_MODEL, notify_function, args1, args2)
            self.thread1.start()
            self.tid = self.thread1.get_my_tid()
        return (None, None)

    def get_model_info(self):
        return (self.model_id, self.reactions, self.metabolites, self.genes)

    def generate_critical_cp_report(self, config):
        logger = CustomLogger(config.print_f, config.args1, config.args2)

        f = FacadeUtils()
        f.processes = config.processes

        results = f.compute_critical_points(
            config.model_path,
            config.print_f,
            config.args1,
            config.args2,
            config.objective,
            config.fraction,
        )
        self.spreadsheet = f.generate_critical_cp_spreadsheet(results)
        if self.spreadsheet is not None and config.output_path_spreadsheet is not None:
            logger.print("Generating spreadsheet report...")
            self.save_spreadsheet(False, config.output_path_spreadsheet, config.print_f)
            _, filename = os.path.split(config.output_path_spreadsheet)
            logger.print("File saved at: " + filename)

        if config.output_path_html is not None:
            logger.print("Generating html report...")
            html = f.generate_critical_cp_html_report(results)
            write_file(config.output_path_html, html)
            _, filename = os.path.split(config.output_path_html)
            logger.print("File saved at: " + filename)

    def generate_spreadsheet(
        self,
        stoppable,
        model_path,
        print_f,
        args1=None,
        args2=None,
        output_path=None,
        objective=None,
        fraction=1.0,
    ):
        warnings.warn(
            "generate_spreadsheet() is deprecated, use generate_critical_cp_report(findCPcore.CriticalCPConfig) instead",
            DeprecationWarning,
        )
        if not stoppable:
            f = FacadeUtils()
            f.processes = self.__processes
            self.spreadsheet = f.run_summary_model(
                model_path, print_f, args1, args2, objective, fraction
            )
            if self.spreadsheet is not None and output_path is not None:
                self.save_spreadsheet(stoppable, output_path, print_f)
            return ""
        else:
            self.thread1 = FacadeThread(self.model_path)
            self.thread1.set_task(
                TASK_SPREADSHEET,
                print_f,
                args1,
                args2,
                output_path,
                objective=objective,
                fraction=fraction,
            )
            self.thread1.start()
            self.tid = self.thread1.get_my_tid()

    def generate_growth_dependent_report(self, config):

        logger = CustomLogger(config.print_f, config.args1, config.args2)

        f = FacadeUtils()
        f.processes = config.processes

        results, state = f.compute_growth_dependent_chokepoints(
            config.model_path,
            config.print_f,
            config.args1,
            config.args2,
            config.objective,
        )
        self.spreadsheet = f.generate_growth_dependent_spreadsheet(results, state)
        if self.spreadsheet is not None and config.output_path_spreadsheet is not None:
            logger.print("Generating spreadsheet report...")
            self.save_spreadsheet(False, config.output_path_spreadsheet, config.print_f)
            _, filename = os.path.split(config.output_path_spreadsheet)
            logger.print("File saved at: " + filename)

        if config.output_path_html is not None:
            logger.print("Generating html report...")
            html = f.generate_growth_dependent_html_report(results)
            write_file(config.output_path_html, html)
            _, filename = os.path.split(config.output_path_html)
            logger.print("File saved at: " + filename)

    def generate_sensibility_spreadsheet(
        self,
        stoppable,
        model_path,
        print_f,
        args1=None,
        args2=None,
        output_path=None,
        objective=None,
    ):
        warnings.warn(
            "generate_sensibility_spreadsheet() is deprecated, use generate_growth_dependent_report(findCPcore.GrowthDependentCPConfig) instead",
            DeprecationWarning,
        )
        if not stoppable:
            f = FacadeUtils()
            f.processes = self.__processes
            self.spreadsheet = f.run_sensibility_analysis(
                model_path, print_f, args1, args2, objective
            )
            if self.spreadsheet is not None and output_path is not None:
                self.save_spreadsheet(stoppable, output_path, print_f)
            return ""
        else:
            self.thread1 = FacadeThread(self.model_path)
            self.thread1.set_task(
                TASK_SENSIBILITY,
                print_f,
                args1,
                args2,
                output_path,
                objective=objective,
            )
            self.thread1.start()
            self.tid = self.thread1.get_my_tid()

    def save_spreadsheet(self, stoppable, output_path, print_f):
        if not stoppable:
            f = FacadeUtils()
            f.processes = self.__processes
            (result_ok, text) = f.save_spreadsheet(output_path, self.spreadsheet)
            return (result_ok, text)
        else:
            self.thread1.set_task(
                TASK_SAVE_SPREADSHEET,
                notify_function=print_f,
                args1=None,
                args2=None,
                output_path=output_path,
            )
            self.thread1.run()
        return (None, None)

    # print_f (arg1, ended=True, str)
    def find_and_remove_dem(
        self, stoppable, output_path, print_f, model_path, args1=None, args2=None
    ):
        if not stoppable:
            f = FacadeUtils()
            f.processes = self.__processes
            self.model = f.find_and_remove_dem(model_path)
        else:
            self.thread1 = FacadeThread(self.model_path)
            self.thread1.set_task(
                TASK_FIND_AND_REMOVE_DEM,
                notify_function=print_f,
                args1=args1,
                args2=None,
                output_path=output_path,
            )
            self.thread1.start()
            self.tid = self.thread1.get_my_tid()

    def run_fva(
        self,
        stoppable,
        output_path,
        print_f,
        model_path,
        args1=None,
        args2=None,
        objective=None,
        fraction=1.0,
    ):
        if not stoppable:
            f = FacadeUtils()
            f.processes = self.__processes
            (self.model, errors) = f.run_fva(model_path, objective, fraction)
            return errors
        else:
            self.thread1 = FacadeThread(self.model_path)
            self.thread1.set_task(
                TASK_SAVE_FVA,
                notify_function=print_f,
                args1=args1,
                args2=None,
                output_path=output_path,
                objective=objective,
                fraction=fraction,
            )
            self.thread1.start()
            self.tid = self.thread1.get_my_tid()

    def run_fva_remove_dem(
        self,
        stoppable,
        output_path,
        print_f,
        model_path,
        args1=None,
        args2=None,
        objective=None,
        fraction=1.0,
    ):
        if not stoppable:
            f = FacadeUtils()
            f.processes = self.__processes
            (self.model, errors) = f.run_fva_remove_dem(model_path, objective, fraction)
            return errors
        else:
            self.thread1 = FacadeThread(self.model_path)
            self.thread1.set_task(
                TASK_SAVE_FVA_DEM,
                notify_function=print_f,
                args1=args1,
                args2=None,
                output_path=output_path,
                objective=objective,
                fraction=fraction,
            )
            self.thread1.start()
            self.tid = self.thread1.get_my_tid()

    # guarda el modelo calculado que se almacena en self
    # print_f (arg, result_ok=True/False, error text)
    def save_model(self, output_path, stoppable, print_f):
        if stoppable:
            self.thread1.set_task(
                TASK_SAVE_MODEL,
                notify_function=print_f,
                args1=None,
                args2=None,
                output_path=output_path,
            )
            self.thread1.run()
        else:
            f = FacadeUtils()
            f.processes = self.__processes
            (result, text) = f.save_model(output_path, self.model)
            return (result, text)



class CustomLogger:
    def __init__(self, print_function, args1, args2):
        self.print_function = print_function
        self.args1 = args1
        self.args2 = args2

    def print(self, string):
        self.print_function(string, self.args1, self.args2)



CELLS_NOT_EQUAL_ERROR = """
Not equal values found on sheet '{}' (row:{}, col:{})

Value found on '{}':
    {}
Value found on '{}':
    {}
"""


class CellsNotEqualException(Exception):
    """Exception raised when two cells on two spreadsheets files have different values.

    Attributes:
        sheet -- sheet name
        row   -- row of the sheet
        col   -- column of the sheet
        file1 -- speadsheet file 1
        val1  -- value on speadsheet file 1 on row, col
        file2 -- speadsheet file 2
        val2  -- value on speadsheet file 2 on row, col
    """

    def __init__(self, sheet, row, col, file1, val1, file2, val2):
        self.sheet = sheet
        self.row = row
        self.col = col
        self.file1 = file1
        self.val1 = val1
        self.file2 = file2
        self.val2 = val2

    def __str__(self):
        return CELLS_NOT_EQUAL_ERROR.format(
            self.sheet, self.row, self.col, self.file1, self.val1, self.file2, self.val2
        )



from __future__ import absolute_import

import click

__version__ = "0.0.0"

from .commands import report, new_model


@click.group()
@click.help_option("--help", "-h")
@click.version_option(__version__, "--version", "-V")
def contrabass():
    """
    Compute vulnerabilities on constrained-based models
    """
    pass


contrabass.add_command(report)
contrabass.add_command(new_model)


if __name__ == "__main__":
    contrabass()



import os
import pytest
import logging
import subprocess
import xlrd
import math
import sys

sys.path.append("src")

from contrabass.core.CobraMetabolicModel import CobraMetabolicModel
from errorHandler import CellsNotEqualException


LOGGER = logging.getLogger(__name__)
LOGGER_MSG = "Executing cli file: {} {}"

# NOTE: these directories are relative the 'src' package
TEST_MODEL = "../test/test_cli/data/MODEL1108160000_url.xml"
INCORRECTLY_FORMATED = "../test/test_cli/data/incorrectly_formatted.xml"
DATA_SPREADSHEET = "../test/test_cli/data/MODEL1108160000_url.xls"
DATA_SPREADSHEET_GROWTH_DEP = "../test/test_cli/data/MODEL1108160000_url_cp.xls"

# directory with the package of the 'contrabass' module
# this directory will be opened with the 'cd' command
CONTRABASS_PACKAGE = os.path.abspath("src")
# this is the whole module of the CLI script
CONTRABASS_MODULE = "contrabass.cli.main"

OUTPUT_MODEL = "/tmp/output.xml"
OUTPUT_MODEL_YML = "/tmp/output.yml"
OUTPUT_SPREADSHEET_TEST = "/tmp/test_spreadsheet.xls"
OUTPUT_SPREADSHEET_TEST_CP = "/tmp/test_spreadsheet_cp.xls"


def __equal_cell(val1, val2):
    if isinstance(val1, float):
        if math.isnan(val1) and math.isnan(val2):
            return True
        else:
            return math.isclose(val1, val2, rel_tol=1e-5)
    else:
        return val1 == val2


def __compare_two_spreadsheets_content(file1, file2):
    xls1 = xlrd.open_workbook(file1, on_demand=True)
    xls2 = xlrd.open_workbook(file1, on_demand=True)

    for name in xls1.sheet_names():
        sheet1 = xls1.sheet_by_name(name)
        sheet2 = xls2.sheet_by_name(name)
        LOGGER.info(
            "Checking sheet : (col:{},row:{}) {} ".format(
                sheet1.ncols, sheet1.nrows, name
            )
        )

        for col in range(sheet1.ncols):
            for row in range(sheet1.nrows):
                val1 = sheet1.cell_value(row, col)
                val2 = sheet2.cell_value(row, col)
                if not __equal_cell(val1, val2):
                    raise CellsNotEqualException(
                        name, row, col, file1, val1, file2, val2
                    )


"""
    Assure correct output and no errors running correct model
"""
@pytest.mark.skip(reason="skipped as it takes to much to run on github action. Running it locally is suggested")
def test_verbose_output_on_model():
    params = [
        "report",
        "critical-reactions",
        TEST_MODEL,
        "--output-spreadsheet",
        OUTPUT_SPREADSHEET_TEST,
    ]

    LOGGER.info(LOGGER_MSG.format(CONTRABASS_PACKAGE, params))

    result = subprocess.check_output(
        ["python", "-m", CONTRABASS_MODULE] + params,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        cwd=CONTRABASS_PACKAGE,
    )

    LOGGER.info(
        "Comparing spreadsheets: '{}' : '{}'".format(
            OUTPUT_SPREADSHEET_TEST, DATA_SPREADSHEET
        )
    )

    __compare_two_spreadsheets_content(OUTPUT_SPREADSHEET_TEST, DATA_SPREADSHEET)


"""
    Assure correct output and no errors running correct model
"""
@pytest.mark.skip(reason="skipped as it takes to much to run on github action. Running it locally is suggested")
def test_verbose_chokepoint_computation_on_model():
    params = [
        "report",
        "growth-dependent-reactions",
        TEST_MODEL,
        "--output-spreadsheet",
        OUTPUT_SPREADSHEET_TEST_CP,
    ]

    LOGGER.info(LOGGER_MSG.format(CONTRABASS_PACKAGE, params))

    result = subprocess.check_output(
        ["python", "-m", CONTRABASS_MODULE] + params,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        cwd=CONTRABASS_PACKAGE,
    )

    LOGGER.info(
        "Comparing spreadsheets: '{}' : '{}'".format(
            OUTPUT_SPREADSHEET_TEST_CP, DATA_SPREADSHEET_GROWTH_DEP
        )
    )

    __compare_two_spreadsheets_content(
        OUTPUT_SPREADSHEET_TEST_CP, DATA_SPREADSHEET_GROWTH_DEP
    )


"""
    Assure correct output and no errors generating new models
"""
@pytest.mark.skip(reason="")
def test_verbose_generate_new_models():
    params = [
        "new-model",
        "fva-constrained-without-dem",
        TEST_MODEL,
        "--output-model",
        OUTPUT_MODEL_YML,
    ]

    LOGGER.info(LOGGER_MSG.format(CONTRABASS_PACKAGE, params))

    result = subprocess.check_output(
        ["python", "-m", CONTRABASS_MODULE] + params,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        cwd=CONTRABASS_PACKAGE,
    )

    model2 = CobraMetabolicModel(OUTPUT_MODEL_YML)
    model2.find_dem()
    for compartment in model2.dem():
        assert len(model2.dem()[compartment]) == 0



from __future__ import absolute_import

__version__ = "0.1.0"

from contrabass.core.CobraMetabolicModel import *
from contrabass.core.State import *
from contrabass.core.Spreadsheet import *
from contrabass.core.FacadeUtils import *
from contrabass.core.Facade import *
from contrabass.core.FacadeThread import *



import logging
import sys
from enum import Enum

sys.path.append("src")

from contrabass.core.CobraMetabolicModel import CobraMetabolicModel
from generate_data import (
    TESTDATA_PRECOMPUTED_DEM,
    TESTDATA_PRECOMPUTED_CHOKEPOINTS,
    TESTDATA_PRECOMPUTED_GROWTH_CP,
)

LOGGER = logging.getLogger(__name__)
LOGGER_MSG = "Executing cli file: {} {}"

TEST_MODEL = "test/test_core/data/MODEL1507180056_url.xml"

EPSILON = 1e-8


class Direction(Enum):
    FORWARD = 0
    BACKWARD = 1
    REVERSIBLE = 2


def reaction_direction(reaction):
    if -EPSILON < reaction.upper_bound < EPSILON:
        upper = 0
    else:
        upper = reaction.upper_bound
    if -EPSILON < reaction.lower_bound < EPSILON:
        lower = 0
    else:
        lower = reaction.lower_bound

    if lower == 0 and upper == 0:
        # Note that reactions with upper and lower bound equal to 0 are considered FORWARD by default
        return Direction.FORWARD
    elif lower >= 0:
        return Direction.FORWARD
    elif upper > 0:
        return Direction.REVERSIBLE
    else:
        return Direction.BACKWARD


def flux_dependent_reactants_products(reaction):
    if reaction_direction(reaction) == Direction.FORWARD:
        reactants1 = reaction.reactants
        products1 = reaction.products
    elif reaction_direction(reaction) == Direction.REVERSIBLE:
        reactants1 = reaction.reactants + reaction.products
        products1 = reaction.reactants + reaction.products
    elif reaction_direction(reaction) == Direction.BACKWARD:
        reactants1 = reaction.products
        products1 = reaction.reactants
    else:
        raise RuntimeError("Error flux_dependent_reactants_products()")

    return reactants1, products1


"""
    Check that dead reactions have flux <closer> to zero.
"""
# @pytest.mark.skip(reason="")
def test_dead_reactions():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    for r in model.dead_reactions():
        assert EPSILON > abs(r.upper_bound)

    model.fva(update_flux=True)

    for r in model.dead_reactions():
        assert EPSILON > abs(r.upper_bound)


"""
    Check dead end metabolites corectness.
"""


def test_dead_end_metabolites():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    model.find_dem()
    model.dem()

    # A reversible reaction cannot produce/consume a DEM
    for c in model.compartments():
        for dem in model.dem()[c]:
            for r in dem.reactions:
                assert reaction_direction(r) != Direction.REVERSIBLE

    # Check they are only produced / consumedç
    for c in model.compartments():
        for dem in model.dem()[c]:
            reactants, products = flux_dependent_reactants_products(
                list(dem.reactions)[0]
            )
            is_product = dem in products

            for r in dem.reactions:
                reactants, products = flux_dependent_reactants_products(r)
                if is_product:
                    assert dem in products
                else:
                    assert dem in reactants


"""
    Check dead end metabolites corectness.
"""


def test_precomputed_dem():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    model.find_dem()
    model.dem()

    sets = set()
    for c in model.compartments():
        for dem in model.dem()[c]:
            sets.add(dem.id)

    # check both sets are identical
    assert len(sets.intersection(TESTDATA_PRECOMPUTED_DEM)) == len(sets)


"""
    Check dead end are removed.
"""


def test_remove_dem():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    model.find_dem()
    model.remove_dem()
    model.find_dem()

    for c in model.compartments():
        assert len(model.dem()[c]) == 0


"""
    Check chokepoint reactions are the only produced/consumer of metabolites
"""


def test_chokepoints():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    dead_reactions = model.dead_reactions()

    model.find_chokepoints()
    for r, m in model.chokepoints():
        counter_produced = 0
        counter_consumed = 0

        for r_test in m.reactions:
            if r_test not in dead_reactions:  # dead reactions are not chokepoints
                reactants, products = flux_dependent_reactants_products(r_test)
                if m in products:
                    counter_produced += 1
                if m in reactants:
                    counter_consumed += 1

                if counter_produced > 1 and counter_consumed > 1:
                    print(m.id + " --- " + r_test)

        assert counter_consumed <= 1 or counter_produced <= 1


"""
    Check pre-computed chokepoint reactions
"""


def test_precomputed_chokepoints():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    cps = set()

    model.find_chokepoints()
    for r, m in model.chokepoints():
        cps.add(r.id)

    # check both sets are identical
    assert len(cps.intersection(TESTDATA_PRECOMPUTED_CHOKEPOINTS)) == len(cps)


"""
    Check pre-computed growth dependent chokepoint reactions
"""


def test_precomputed_growth_chokepoints():
    model = CobraMetabolicModel(TEST_MODEL)
    EPSILON = model.epsilon()

    model.fva(update_flux=True)

    cps = set()

    model.find_chokepoints()
    for r, m in model.chokepoints():
        cps.add(r.id)

    # check both sets are identical
    assert len(cps.intersection(TESTDATA_PRECOMPUTED_GROWTH_CP)) == len(cps)



import itertools
import json
from json import JSONEncoder

from contrabass.core.report_utils import simplified_reactions, simplified_metabolites, simplified_genes, format_annotation


class Reaction:
    id = None
    name = None
    reaction = None
    upper_bound = None
    lower_bound = None
    gene_reaction_rule = None
    metabolites = None
    genes = None
    annotation = None

    def __init__(
        self, id, name, reaction, upper_bound, lower_bound, gene_reaction_rule, annotation
    ):
        self.id = id
        self.name = name
        self.reaction = reaction
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound
        self.gene_reaction_rule = gene_reaction_rule
        self.metabolites = []
        self.genes = []
        self.annotation = annotation

    def add_metabolite(self, metabolite):
        self.metabolites.append(metabolite)

    def add_gene(self, genes):
        self.genes.append(genes)


class Metabolite:
    id = None
    name = None
    compartment = None
    reactions = None
    annotation = None
    formula = None

    def __init__(self, id, name, compartment, annotation, formula):
        self.id = id
        self.name = name
        self.compartment = compartment
        self.reactions = []
        self.annotation = annotation
        self.formula = formula

    def add_reaction(self, reaction):
        self.reactions.append(reaction)


class Gene:
    id = None
    name = None
    reactions = None
    annotation = None

    def __init__(self, id, name, annotation):
        self.id = id
        self.name = name
        self.reactions = []
        self.annotation = annotation

    def add_reaction(self, reaction):
        self.reactions.append(reaction)


class State:
    __id = None
    __objective = None
    __objective_value = None

    __reactions = None
    __metabolites = None
    __genes = None

    __dem = None
    __chokepoints = None
    __fva = None
    __essential_genes = None
    __essential_genes_reactions = None
    __knockout_growth = None
    __essential_reactions = None

    __error = None

    def id(self):
        return self.__id

    def set_id(self, id):
        self.__id = id

    def objective(self):
        return self.__objective

    def set_objective(self, objective):
        self.__objective = objective

    def objective_value(self):
        return self.__objective_value

    def set_objective_value(self, objective_value):
        self.__objective_value = objective_value

    def reactions(self):
        return self.__reactions

    def set_reactions(self, reactions):
        self.__reactions = reactions

    def metabolites(self):
        return self.__metabolites

    def set_metabolites(self, metabolites):
        self.__metabolites = metabolites

    def genes(self):
        return self.__genes

    def set_genes(self, genes):
        self.__genes = genes

    def dem(self):
        return self.__dem

    def set_dem(self, dem):
        self.__dem = dem

    def chokepoints(self):
        return self.__chokepoints

    def set_chokepoints(self, chokepoints):
        self.__chokepoints = chokepoints

    def fva(self):
        return self.__fva

    def set_fva(self, fva):
        self.__fva = fva

    def essential_genes(self):
        return self.__essential_genes

    def set_essential_genes(self, essential_genes):
        self.__essential_genes = essential_genes

    def essential_genes_reactions(self):
        return self.__essential_genes_reactions

    def set_essential_genes_reactions(self, essential_genes_reactions):
        self.__essential_genes_reactions = essential_genes_reactions

    def knockout_growth(self):
        return self.__knockout_growth

    def set_knockout_growth(self, knockout_growth):
        self.__knockout_growth = knockout_growth

    def essential_reactions(self):
        return self.__essential_reactions

    def set_essential_reactions(self, essential_reactions):
        self.__essential_reactions = essential_reactions

    def reversible_reactions(self):
        return self.__reversible_reactions

    def set_reversible_reactions(self, reversible_reactions):
        self.__reversible_reactions = reversible_reactions

    def dead_reactions(self):
        return self.__dead_reactions

    def set_dead_reactions(self, dead_reactions):
        self.__dead_reactions = dead_reactions

    def set_error(self, error):
        self.__error = error

    def error(self):
        return self.__error

    def to_json(self):

        return {
            "id": self.__id,
            "objective": self.__objective,
            "objective_value": self.__objective_value,
            "reactions": simplified_reactions(self.__reactions),
            "metabolites": simplified_metabolites(self.__metabolites),
            "genes": simplified_genes(self.__genes),
            "dem": [
                m.id for m in list(itertools.chain.from_iterable(self.__dem.values()))
            ],
            "chokepoints": [(r.id, m.id) for (r, m) in self.__chokepoints],
            "fva": [
                (reaction.id, upper, lower) for (reaction, upper, lower) in self.__fva
            ],
            "essential_genes": [g.id for g in self.__essential_genes],
            "essential_genes_reactions": [
                r.id for r in self.__essential_genes_reactions
            ],
            "knockout_growth": [(r.id, g) for r, g in self.__knockout_growth.items()],
            "essential_reactions": self.__essential_reactions,
            "dead_reactions": self.__dead_reactions,
            "reversible_reactions": self.__reversible_reactions,
            "error": self.__error if self.__error is not None else "",
        }


class CobraMetabolicStateBuilder:

    reactions = None
    metabolites = None
    genes = None

    def buildState(self, model, error=None):
        state = State()
        state.set_id(self.id(model))
        state.set_objective(self.objective(model))
        state.set_objective_value(self.objective_value(model))
        state.set_reversible_reactions(self.reversible_reactions(model))
        state.set_dead_reactions(self.dead_reactions(model))
        (metabolites, reactions, genes) = self.metabolic_network(model)
        state.set_reactions(reactions)
        state.set_metabolites(metabolites)
        state.set_genes(genes)
        state.set_essential_reactions(self.essential_reactions(model))
        state.set_dead_reactions(self.dead_reactions(model))
        state.set_reversible_reactions(self.reversible_reactions(model))
        state.set_error(error)
        try:
            state.set_dem(self.dem(model))
        except Exception:
            state.set_dem({})
        try:
            state.set_chokepoints(self.chokepoints(model))
        except Exception:
            state.set_chokepoints([])
        try:
            state.set_fva(self.fva(model))
        except Exception:
            state.set_fva([])
        try:
            state.set_essential_genes(self.essential_genes(model))
        except Exception:
            state.set_essential_genes([])
        try:
            state.set_essential_genes_reactions(self.essential_genes_reactions(model))
        except Exception:
            state.set_essential_genes_reactions({})
        try:
            state.set_knockout_growth(self.knockout_growth(model))
        except Exception:
            state.set_knockout_growth({})

        return state

    def id(self, model):
        return model.id()

    def objective(self, model):
        return model.objective()

    def objective_value(self, model):
        if model.objective_value() is None:
            return None
        else:
            return model.objective_value()

    def metabolic_network(self, model):
        metabolites = {}
        for metabolite in model.metabolites():
            metabolites[metabolite.id] = Metabolite(
                metabolite.id, metabolite.name, metabolite.compartment, metabolite.annotation, metabolite.formula
            )
        reactions = {}
        for reaction in model.reactions():
            reactions[reaction.id] = Reaction(
                reaction.id,
                reaction.name,
                reaction.reaction,
                reaction.upper_bound,
                reaction.lower_bound,
                reaction.gene_reaction_rule,
                reaction.annotation
            )
        genes = {}
        for gene in model.genes():
            genes[gene.id] = Gene(gene.id, gene.name, gene.annotation)

        self.metabolites = metabolites
        self.reactions = reactions
        self.genes = genes

        for reaction in model.reactions():
            new_reaction = reactions[reaction.id]
            for gene in reaction.genes:
                new_gene = genes[gene.id]
                new_reaction.add_gene(new_gene)
        for gene in model.genes():
            new_gene = genes[gene.id]
            for reaction in gene.reactions:
                new_reaction = reactions[reaction.id]
                new_gene.add_reaction(new_reaction)
        for metabolite in model.metabolites():
            new_metabolite = metabolites[metabolite.id]
            for reaction in metabolite.reactions:
                new_reaction = reactions[reaction.id]
                new_metabolite.add_reaction(new_reaction)

        return (
            list(metabolites.values()),
            list(reactions.values()),
            list(genes.values()),
        )

    def dem(self, model):
        result = None
        dem = model.dem()
        if dem is not None:
            result = {}
            for compartment in dem.keys():
                list_c = [
                    self.metabolites[metabolite.id] for metabolite in dem[compartment]
                ]
                result[compartment] = list_c
        return result

    def chokepoints(self, model):
        result = None
        chokepoints = model.chokepoints()
        if chokepoints is not None:
            result = []
            for (reaction, metabolite) in chokepoints:
                result.append(
                    (self.reactions[reaction.id], self.metabolites[metabolite.id])
                )
        return result

    def fva(self, model):
        try:
            result = []
            model_fva = model.get_fva()
            if model_fva is not None:
                for el in model_fva:
                    (reaction, upper, lower) = el
                    local_reaction = self.reactions[reaction.id]
                    result.append((local_reaction, upper, lower))
            return result
        except Exception:
            return []

    def essential_genes(self, model):
        try:
            result = []
            genes = model.essential_genes()
            if genes is not None:
                gen_list = [self.genes[gen.id] for gen in genes]
                result = gen_list
            return result
        except Exception:
            return []

    def essential_genes_reactions(self, model):
        try:
            result = {}
            egr = model.essential_genes_reactions()
            if egr is not None:
                result = {}
                for reaction in egr:
                    for gen in egr[reaction]:
                        if reaction.id not in result:
                            result[self.reactions[reaction.id]] = [self.genes[gen.id]]
                        else:
                            result[self.reactions[reaction.id]].append(
                                self.genes[gen.id]
                            )
            return result
        except Exception:
            return {}

    def knockout_growth(self, model):
        try:
            result = None
            essr = model.knockout_growth()
            if essr is not None:
                result = {}
                for reaction in essr.keys():
                    result[self.reactions[reaction.id]] = essr[reaction]
            return result
        except Exception:
            return {}

    def essential_reactions(self, model):
        model.compute_essential_reactions()
        return [r.id for r in model.essential_reactions()]

    def reversible_reactions(self, model):
        return [r.id for r in model.reversible_reactions()]

    def dead_reactions(self, model):
        return [r.id for r in model.dead_reactions()]


class StateEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, State):
            return o.to_json()
        return json.JSONEncoder.default(self, o)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sphinx_rtd_theme
# Use sphinx-quickstart to create your own conf.py file!
# After that, you have to edit a few things.  See below.

# Select nbsphinx and, if needed, add a math extension (mathjax or pngmath):
extensions = [
    'nbsphinx',
    'sphinx.ext.mathjax',
    'recommonmark'
]

# for Sphinx-1.3
from recommonmark.parser import CommonMarkParser

source_parsers = {
    '.md': CommonMarkParser,
}

source_suffix = ['.rst', '.md']

# Exclude build directory and Jupyter backup files:
exclude_patterns = ['_build', '**.ipynb_checkpoints']

# Default language for syntax highlighting in reST and Markdown cells
highlight_language = 'none'

# Don't add .txt suffix to source files (available for Sphinx >= 1.5):
html_sourcelink_suffix = ''

# Execute notebooks before conversion: 'always', 'never', 'auto' (default)
nbsphinx_execute = 'never'

# Use this kernel instead of the one stored in the notebook metadata:
nbsphinx_kernel_name = 'python3'

# List of arguments to be passed to the kernel that executes the notebooks:
#nbsphinx_execute_arguments = ['--InlineBackend.figure_formats={"png", "pdf"}']

# If True, the build process is continued even if an exception occurs:
nbsphinx_allow_errors = True

# Controls when a cell will time out (defaults to 30; use -1 for no timeout):
#nbsphinx_timeout = 60

# Default Pygments lexer for syntax highlighting in code cells:
#nbsphinx_codecell_lexer = 'ipython3'

# Width of input/output prompts used in CSS:
#nbsphinx_prompt_width = '8ex'

#import stanford_theme
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
#html_theme_path = [stanford_theme.get_html_theme_path()]
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# If window is narrower than this, input/output prompts are on separate lines:
#nbsphinx_responsive_width = '700px'

# -- The settings below this line are not specific to nbsphinx ------------

master_doc = 'index'

project = 'contrabass'
author = 'Alex Oarga'
copyright = '2021, ' + author + " <718123 at unizar dot es>"

linkcheck_ignore = [r'http://localhost:\d+/']

# -- Get version information from Git -------------------------------------

try:
    from subprocess import check_output
    release = check_output(['git', 'describe', '--tags', '--always'])
    release = release.decode().strip()
except Exception:
    release = '<unknown>'

# -- Options for HTML output ----------------------------------------------

html_title = project + ' version ' + release

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    'papersize': 'a4paper',
    'preamble': r"""
\usepackage[sc,osf]{mathpazo}
\linespread{1.05}  % see http://www.tug.dk/FontCatalogue/urwpalladio/
\renewcommand{\sfdefault}{pplj}  % Palatino instead of sans serif
\IfFileExists{zlmtt.sty}{
    \usepackage[light,scaled=1.05]{zlmtt}  % light typewriter font from lmodern
}{
    \renewcommand{\ttdefault}{lmtt}  % typewriter font from lmodern
}
""",
}

latex_documents = [
    (master_doc, 'nbsphinx.tex', project, author, 'howto'),
]

latex_show_urls = 'footnote'



import os
import pytest
import logging
import subprocess
import xlrd
import math
import sys
from dotenv import load_dotenv

load_dotenv()

ENV_ENVIRONMENT = "ENVIRONMENT"
ENV_DEV = "DEV"
ENV_PRO = "PRO"

if os.environ.get(ENV_ENVIRONMENT) == ENV_DEV:
    # import relative path of core for development purposes
    print("*** DEVELOPMENT ENVIRONMENT ***")
    sys.path.append("../../findCPcore_pkg/core/")
    from CobraMetabolicModel import CobraMetabolicModel
else:
    from findCPcore import CobraMetabolicModel

from errorHandler import CellsNotEqualException


LOGGER = logging.getLogger(__name__)
LOGGER_MSG = "Executing cli file: {} {}"

TEST_MODEL = "data/MODEL1108160000_url.xml"
INCORRECTLY_FORMATED = "data/incorrectly_formatted.xml"
FINDCPCLI_PATH = os.path.abspath("../../scripts/contrabass")

OUTPUT_SPREADSHEET_TEST = "data/test_spreadsheet.xls"
OUTPUT_SPREADSHEET_TEST_CP = "data/test_spreadsheet_cp.xls"


def generate_test_no_args():

    LOGGER.info("Executing cli file: {}".format(FINDCPCLI_PATH))

    try:
        subprocess.check_output(
            ["python", FINDCPCLI_PATH], stderr=subprocess.STDOUT, text=True
        )
        raise RuntimeError(
            "This point should not be reached. Return code must not be 0."
        )
    except subprocess.CalledProcessError as err:
        f = open("data/no_args_err.txt", "w")
        f.write(str(err.stdout))
        f.close()


def generate_test_incorrect_input_model():

    try:
        params = ["-i", INCORRECTLY_FORMATED, "-o", OUTPUT_SPREADSHEET_TEST]

        LOGGER.info(LOGGER_MSG.format(FINDCPCLI_PATH, params))

        result = subprocess.check_output(
            ["python", FINDCPCLI_PATH] + params, stderr=subprocess.STDOUT, text=True
        )

    except subprocess.CalledProcessError as err:
        f = open("data/test_incorrect_input_model.txt", "w")
        f.write(str(err.stdout))
        f.close()


def generate_test_verbose_output_on_model():
    params = ["-i", TEST_MODEL, "-o", OUTPUT_SPREADSHEET_TEST, "-v"]

    LOGGER.info(LOGGER_MSG.format(FINDCPCLI_PATH, params))

    result = ""
    try:
        result = subprocess.check_output(
            ["python", FINDCPCLI_PATH] + params, stderr=subprocess.STDOUT, text=True
        )
    except subprocess.CalledProcessError as err:
        print(err.stdout)

    f = open("data/test_verbose_output_on_model.txt", "w")
    f.write(str(result))
    f.close()


def generate_test_verbose_chokepoint_computation_on_model():
    params = ["-i", TEST_MODEL, "-cp", OUTPUT_SPREADSHEET_TEST_CP, "-v"]

    LOGGER.info(LOGGER_MSG.format(FINDCPCLI_PATH, params))

    result = subprocess.check_output(
        ["python", FINDCPCLI_PATH] + params, stderr=subprocess.STDOUT, text=True
    )

    f = open("data/test_verbose_chokepoint_computation_on_model.txt", "w")
    f.write(str(result))
    f.close()


def generate_test_verbose_generate_new_models():
    params = [
        "-i",
        TEST_MODEL,
        "-swD",
        "data/model_without_dem.xml",
        "-sF",
        "data/model_fva_flux.json",
        "-swDF",
        "data/model_fva_dem.yml",
    ]

    LOGGER.info(LOGGER_MSG.format(FINDCPCLI_PATH, params))

    result = subprocess.check_output(
        ["python", FINDCPCLI_PATH] + params, stderr=subprocess.STDOUT, text=True
    )


generate_test_no_args()
generate_test_incorrect_input_model()
generate_test_verbose_output_on_model()
generate_test_verbose_chokepoint_computation_on_model()
generate_test_verbose_generate_new_models()



from copy import copy

import xlwt
import itertools
from math import isnan


class Spreadsheet:

    __spreadsheet = None

    def spreadsheet_init(self):
        """Initialize the xlwt Workbook"""
        self.__spreadsheet = xlwt.Workbook()

    def set_workbook(self, workbook):
        self.__spreadsheet = workbook

    def get_workbook(self):
        return self.__spreadsheet

    def spreadsheet_write_model_info(self, state, sheet_name):
        """Writes in a sheet of the xlwt Workbook '__spreadsheet' info of the model with the following format:

                        MODEL_ID | REACTIONS | METABOLITES | GENES
                        ---------+-----------+-------------+-------
                        model id | len(reac.)| len(metab.) | len(genes)

        Args:
                sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "MODEL_ID", style=style)
        sheet.write(0, 1, state.id())
        sheet.write(1, 0, "REACTIONS", style=style)
        sheet.write(1, 1, len(state.reactions()))
        sheet.write(2, 0, "METABOLITES", style=style)
        sheet.write(2, 1, len(state.metabolites()))
        sheet.write(3, 0, "GENES", style=style)
        sheet.write(3, 1, len(state.genes()))
        sheet.write(4, 0, "OBJECTIVE FUNCTION", style=style)
        sheet.write(4, 1, state.objective())

    def __id(self, e):
        return e.id

    def __spreadsheet_write_metabolites(
        self, list, sheet_name, ordered=True, print_reactions=False
    ):
        """Writes in a sheet of the xlwt Workbook '__spreadsheet' the metabolites of the model with the following format:

                METABOLITE ID | METABOLITE NAME | METABOLITE COMPARTMENT | REACTION ID | REACTION NAME | REACTION
                --------------+-----------------+------------------------+-------------+---------------+----------
                mtb 1         | metabolite 1    | compartment 1          | react1      | reaction 1    | mtb 1 -> mtb 2
                --------------+-----------------+------------------------+-------------+---------------+----------
                                          |                 |                        | react2      | reaction 2    | mtb 1 -> mtb 3

        Args:
                list (): List of cobra.core.metabolites to be written.
                sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
                ordered (): Write the metabolites in alphabetical order by id.
                print_reactions (): Write the reactions in which a metabolite appears.
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()
        metabolites = list
        if ordered:
            metabolites.sort(key=self.__id)

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "METABOLITE ID", style=style)
        sheet.write(0, 1, "METABOLITE NAME", style=style)
        sheet.write(0, 2, "METABOLITE COMPARTMENT", style=style)
        if print_reactions:
            sheet.write(0, 3, "REACTION ID", style=style)
            sheet.write(0, 4, "REACTION NAME", style=style)
            sheet.write(0, 5, "UPPER BOUND", style=style)
            sheet.write(0, 6, "LOWER BOUND", style=style)
            sheet.write(0, 7, "REACTION", style=style)
        i = 1
        for metabolite in metabolites:
            sheet.write(i, 0, metabolite.id)
            sheet.write(i, 1, metabolite.name)
            sheet.write(i, 2, metabolite.compartment)
            if print_reactions:
                reactions = []
                for r in metabolite.reactions:
                    reactions.append(r)
                reactions.sort(key=self.__id)
                for reaction in reactions:
                    sheet.write(i, 3, reaction.id)
                    sheet.write(i, 4, reaction.name)
                    sheet.write(i, 5, reaction.upper_bound)
                    sheet.write(i, 6, reaction.lower_bound)
                    sheet.write(i, 7, reaction.reaction)
                    i = i + 1
            if not print_reactions or len(reactions) == 0:
                i = i + 1

    def spreadsheet_write_reactions(
        self, state, sheet_name, ordered=True, print_genes=False
    ):
        """Writes in a sheet of the xlwt Workbook '__spreadsheet' the reactions of the model with the following format:

                REACTION ID | REACTION NAME | REACTION | UPPER BOUND | LOWER BOUND | CHOKEPOINT | ESSENTIAL GENE
                ------------+---------------+----------+-------------+-------------+------------+----------------
                react 1     | reaction 1    | mt1->mt2 | 1000        | 0           | TRUE       |
                ------------+---------------+----------+-------------+-------------+------------+----------------
                react 2     | reaction 2    | mt3->mt4 | 2000        | 0           |            | TRUE

        Args:
                sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
                ordered (): Print the reactions in alphabetical order by id
                tag_chokepoints (): mark a reaction with 'TRUE' if chokepoint reaction
                tag_essential_genes (): mark a reactions with 'TRUE' if associated to an essential gene
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        reactions = state.reactions()
        if ordered:
            reactions.sort(key=self.__id)

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "REACTION ID", style=style)
        sheet.write(0, 1, "REACTION NAME", style=style)
        sheet.write(0, 2, "REACTION GPR", style=style)
        sheet.write(0, 3, "REACTION", style=style)
        sheet.write(0, 4, "UPPER BOUND", style=style)
        sheet.write(0, 5, "LOWER BOUND", style=style)
        i = 1
        for reaction in reactions:
            sheet.write(i, 0, reaction.id)
            sheet.write(i, 1, reaction.name)
            sheet.write(i, 2, reaction.gene_reaction_rule[:32760])
            sheet.write(i, 3, reaction.reaction)
            sheet.write(i, 4, reaction.upper_bound)
            sheet.write(i, 5, reaction.lower_bound)
            i = i + 1

    def spreadsheet_write_metabolites(
        self, state, sheet_name, ordered=True, print_reactions=False
    ):
        """Writes the metabolites of the model with the function '__spreadsheet_write_metabolites'"""
        self.__spreadsheet_write_metabolites(
            state.metabolites(), sheet_name, ordered, print_reactions
        )

    def spreadsheet_write_genes(
        self, state, sheet_name, ordered=False, print_reactions=False
    ):
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        genes = state.genes()
        if ordered:
            genes.sort(key=self.__id)

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "GENE ID", style=style)
        sheet.write(0, 1, "GENE NAME", style=style)
        if print_reactions:
            sheet.write(0, 2, "GENE RELATION", style=style)
            sheet.write(0, 3, "REACTION ID", style=style)
            sheet.write(0, 4, "REACTION NAME", style=style)
        i = 1
        for gen in genes:
            sheet.write(i, 0, gen.id)
            sheet.write(i, 1, gen.name)
            if print_reactions:
                reactions = []
                for r in gen.reactions:
                    reactions.append(r)
                reactions.sort(key=self.__id)
                for reaction in reactions:
                    gpr = (
                        (reaction.gene_reaction_rule[:32760] + "...")
                        if len(reaction.gene_reaction_rule) > 32765
                        else reaction.gene_reaction_rule
                    )
                    sheet.write(i, 2, gpr)
                    sheet.write(i, 3, reaction.id)
                    sheet.write(i, 4, reaction.name)
                    i = i + 1
            if not print_reactions or len(reactions) == 0:
                i = i + 1

    def spreadsheet_write_dem(
        self, state, sheet_name, ordered=False, print_reactions=False, compartment="ALL"
    ):
        """Writes the dead end metabolites of the model with the function '__spreadsheet_write_metabolites'"""
        if state.dem() is not None:
            if compartment == "ALL":
                all_dems = list(itertools.chain.from_iterable(state.dem().values()))
                self.__spreadsheet_write_metabolites(
                    all_dems, sheet_name, ordered, print_reactions
                )
            else:
                if compartment in state.dem().keys():
                    self.__spreadsheet_write_metabolites(
                        state.dem()[compartment], sheet_name, ordered, print_reactions
                    )

    def __reaction_id(self, obj):
        """Given an object _MetaboliteReact it returns the 'id' of the reaction

        Args:
                obj (): Object cobra.core.reaction

        Returns: 'id' field of the reaction

        """
        (reaction, metabolite) = obj
        return reaction.id

    def spreadsheet_write_chokepoints(self, state, sheet_name, ordered=False):
        """"""
        if self.__spreadsheet is None:
            self.spreadsheet_init()
        if state.chokepoints() is not None:
            chokepoints = state.chokepoints()
            if ordered:
                chokepoints.sort(key=self.__reaction_id)
            style = xlwt.XFStyle()
            font = xlwt.Font()
            font.bold = True
            style.font = font
            sheet = self.__spreadsheet.add_sheet(sheet_name)
            sheet.write(0, 0, "REACTION ID", style=style)
            sheet.write(0, 1, "METABOLITE ID", style=style)
            i = 1
            for (reaction, metabolite) in chokepoints:
                sheet.write(i, 0, reaction.id)
                sheet.write(i, 1, metabolite.id)
                i = i + 1

    def spreadsheet_write_reversible_reactions(
        self, sheet_name, state_initial, state_fva, ordered=False
    ):
        """"""
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        reactions = state_initial.reactions()
        if ordered:
            reactions.sort(key=self.__id)

        reversible_initial = state_initial.reversible_reactions()
        reversible_fva = state_fva.reversible_reactions()

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "REACTION ID", style=style)
        sheet.write(0, 1, "REVERSIBLE INITIAL", style=style)
        sheet.write(0, 2, "REVERSIBLE FVA", style=style)
        i = 1
        for reaction in reactions:
            sheet.write(i, 0, reaction.id)
            if reaction.id in reversible_initial:
                sheet.write(i, 1, "TRUE")
            if reaction.id in reversible_fva:
                sheet.write(i, 2, "TRUE")
            i = i + 1

    def __reaction_id_fva(self, obj):
        (react, max, min) = obj
        return react.id

    def spreadsheet_write_fva(self, state, sheet_name, ordered=False):
        """Writes in a sheet of the xlwt Workbook '__spreadsheet' the result of the Flux Variability Analysis.

                REACTION ID | REACTION NAME | REACTION | MAXIMUN | MINIMUM
                ------------+---------------+----------+---------+------------
                react 1     | reaction 1    | mt1->mt2 | 1000    | 0
                ------------+---------------+----------+---------+------------
                react 2     | reaction 2    | mt3->mt4 | 2000    |

        Args:
                sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
                ordered (): Print the reactions in alphabetical order by id
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        if state.fva() is not None:
            fva = state.fva()
            if ordered:
                fva.sort(key=self.__reaction_id_fva)
            style = xlwt.XFStyle()
            font = xlwt.Font()
            font.bold = True
            style.font = font
            sheet = self.__spreadsheet.add_sheet(sheet_name)
            sheet.write(0, 0, "REACTION ID", style=style)
            sheet.write(0, 1, "REACTION NAME", style=style)
            sheet.write(0, 2, "MAX", style=style)
            sheet.write(0, 3, "MIN", style=style)
            i = 1
            for obj in fva:
                (reaction, max, min) = obj
                sheet.write(i, 0, reaction.id)
                sheet.write(i, 1, reaction.name)
                sheet.write(i, 2, max)
                sheet.write(i, 3, min)
                i = i + 1

    def spreadsheet_write_essential_genes(self, state, sheet_name, ordered=False):
        """If the model has genes, writes in a sheet of the xlwt Workbook '__spreadsheet'
                the essential genes of the model with the following format:

                GENE ID | REACTION ID | REACTION NAME | REACTION
                --------+-------------+---------------+-------------
                gene 1  | react 1     | reaction 1    | mt1 + mt2 -> mt3
                --------+-------------+---------------+-------------
                        | react 2     | reaction 2    | mt4 -> mt5
                --------+-------------+---------------+-------------
                gene 2  | react 1     | reaction 1    | mt1 + mt2 -> mt3

        Args:
                sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
                ordered (): Print the genes in alphabetical order by id
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        if state.essential_genes() is not None:

            style = xlwt.XFStyle()
            font = xlwt.Font()
            font.bold = True
            style.font = font
            sheet = self.__spreadsheet.add_sheet(sheet_name)
            sheet.write(0, 0, "GENE ID", style=style)
            sheet.write(0, 1, "GENE NAME", style=style)
            sheet.write(0, 2, "REACTION GENE RELATION", style=style)
            sheet.write(0, 3, "REACTION ID", style=style)
            sheet.write(0, 4, "REACTION NAME", style=style)

            genes = state.essential_genes()
            if ordered:
                genes.sort(key=self.__id)
            row = 1
            for gene in genes:
                reactions = gene.reactions
                sheet.write(row, 0, gene.id)
                sheet.write(row, 1, gene.name)
                gpr = (
                    (reactions[0].gene_reaction_rule[:32760] + "...")
                    if len(reactions[0].gene_reaction_rule) > 32765
                    else reactions[0].gene_reaction_rule
                )
                sheet.write(row, 2, reactions[0].gene_reaction_rule)
                sheet.write(row, 3, reactions[0].id)
                sheet.write(row, 4, reactions[0].name)
                row = row + 1
                i = 1
                while i < len(reactions):
                    gpr = (
                        (reactions[i].gene_reaction_rule[:32760] + "...")
                        if len(reactions[i].gene_reaction_rule) > 32765
                        else reactions[i].gene_reaction_rule
                    )
                    sheet.write(row, 2, gpr)
                    sheet.write(row, 3, reactions[i].id)
                    sheet.write(row, 4, reactions[i].name)
                    i = i + 1
                    row = row + 1

    def spreadsheet_write_essential_genes_reactions(
        self, state, sheet_name, ordered=False
    ):
        """If the model has genes, writes in a sheet of the xlwt Workbook '__spreadsheet'
                the reactions associated to the essential genes of the model with the following format:

                REACTION ID | REACTION NAME | REACTION | GENE ID
                ------------+---------------+----------+-------------
                react 1     | reaction 1    | mt1->mt2 | gene1
                ------------+---------------+----------+-------------
                            |               |          | gene2
                ------------+---------------+----------+-------------
                react 2     | reaction 2    | mt1->mt2 | gene3

        Args:
                sheet_name (): Name of the new sheet. Two diffetent sheets can't have the same name.
                ordered (): Print the reactions in alphabetical order by id
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()
        if state.essential_genes_reactions() is not None:
            reactions = state.essential_genes_reactions()

            style = xlwt.XFStyle()
            font = xlwt.Font()
            font.bold = True
            style.font = font
            sheet = self.__spreadsheet.add_sheet(sheet_name)
            sheet.write(0, 0, "REACTION ID", style=style)
            sheet.write(0, 1, "REACTION NAME", style=style)
            sheet.write(0, 2, "GENE REACTION RELATION", style=style)
            sheet.write(0, 3, "GENE ID", style=style)
            sheet.write(0, 4, "GENE NAME", style=style)
            reactions_key = reactions.keys()
            if ordered:
                reactions_key = sorted(reactions.keys(), key=self.__id)
            row = 1
            for reaction in reactions_key:
                genes = reactions[reaction]
                sheet.write(row, 0, reaction.id)
                sheet.write(row, 1, reaction.name)
                gpr = (
                    (reaction.gene_reaction_rule[:32760] + "...")
                    if len(reaction.gene_reaction_rule) > 32765
                    else reaction.gene_reaction_rule
                )
                sheet.write(row, 2, gpr)
                sheet.write(row, 3, genes[0].id)
                sheet.write(row, 4, genes[0].name)
                row = row + 1
                i = 1
                while i < len(genes):
                    sheet.write(row, 3, genes[i].id)
                    sheet.write(row, 4, genes[i].name)
                    i = i + 1
                    row = row + 1

    def spreadsheet_write_essential_reactions(
        self,
        sheet_name,
        state_initial,
        state_dem,
        state_fva,
        state_fva_dem,
        ordered=True,
    ):
        """"""
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        reactions_initial = {}
        if state_initial.knockout_growth() is not None:
            for reaction, growth in state_initial.knockout_growth().items():
                reactions_initial[reaction.id] = growth
        reactions_dem = {}
        if state_dem.knockout_growth() is not None:
            for reaction, growth in state_dem.knockout_growth().items():
                reactions_dem[reaction.id] = growth
        reactions_fva = {}
        if state_fva.knockout_growth() is not None:
            for reaction, growth in state_fva.knockout_growth().items():
                reactions_fva[reaction.id] = growth
        reactions_fva_dem = {}
        if state_fva_dem.knockout_growth() is not None:
            for reaction, growth in state_fva_dem.knockout_growth().items():
                reactions_fva_dem[reaction.id] = growth

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "REACTION ID", style=style)
        sheet.write(0, 1, "REACTION NAME", style=style)
        sheet.write(0, 2, "OBJECTIVE INITIAL", style=style)
        sheet.write(0, 3, "OBJECTIVE DEM", style=style)
        sheet.write(0, 4, "OBJECTIVE FVA", style=style)
        sheet.write(0, 5, "OBJECTIVE FVA DEM", style=style)
        reactions_initial_key = reactions_initial.keys()
        reactions_dem_key = reactions_dem.keys()
        reactions_fva_key = reactions_fva.keys()
        reactions_fva_dem_key = reactions_fva_dem.keys()
        if ordered:
            reactions = sorted(state_initial.knockout_growth().keys(), key=self.__id)
        row = 1
        for reaction in reactions:
            sheet.write(row, 0, reaction.id)
            sheet.write(row, 1, reaction.name)
            if reaction.id in reactions_initial_key:
                sheet.write(row, 2, reactions_initial[reaction.id])
            if reaction.id in reactions_dem_key:
                sheet.write(row, 3, reactions_dem[reaction.id])
            if reaction.id in reactions_fva_key:
                sheet.write(row, 4, reactions_fva[reaction.id])
            if reaction.id in reactions_fva_dem_key:
                sheet.write(row, 5, reactions_fva_dem[reaction.id])
            row = row + 1

    def spreadsheet_write_essential_genes_comparison(
        self,
        sheet_name,
        state_initial,
        state_dem,
        state_fva,
        state_fva_dem,
        ordered=True,
    ):
        """"""
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        essential_genes_initial = [g.id for g in state_initial.essential_genes()]
        essential_genes_dem = [g.id for g in state_dem.essential_genes()]
        essential_genes_fva = [g.id for g in state_fva.essential_genes()]
        essential_genes_fva_dem = [g.id for g in state_fva_dem.essential_genes()]

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "GENE ID", style=style)
        sheet.write(0, 1, "GENE NAME", style=style)
        sheet.write(0, 2, "MODEL INITIAL", style=style)
        sheet.write(0, 3, "MODEL WITHOUT DEM", style=style)
        sheet.write(0, 4, "MODEL FVA", style=style)
        sheet.write(0, 5, "MODEL FVA WITHOUT DEM", style=style)

        if ordered:
            genes = sorted(state_initial.genes(), key=self.__id)
        else:
            genes = state_initial.genes()

        row = 1
        for gene in genes:
            sheet.write(row, 0, gene.id)
            sheet.write(row, 1, gene.name)
            if gene.id in essential_genes_initial:
                sheet.write(row, 2, "TRUE")
            if gene.id in essential_genes_dem:
                sheet.write(row, 3, "TRUE")
            if gene.id in essential_genes_fva:
                sheet.write(row, 4, "TRUE")
            if gene.id in essential_genes_fva_dem:
                sheet.write(row, 5, "TRUE")
            row = row + 1

    def spreadsheet_write_summary_reactions(
        self, sheet_name, state_initial, state_dem, state_fva, state_fva_dem
    ):
        """

        Args:
                sheet_name ():

        Returns:

        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        reactions = state_initial.reactions()
        reactions.sort(key=self.__id)

        #
        initial_chokepoints_dict = {}
        if state_initial.chokepoints() is not None:
            for el in state_initial.chokepoints():
                (reaction, metabolite) = el
                reaction = reaction.id
                metabolite = metabolite.id
                if reaction in initial_chokepoints_dict:
                    initial_chokepoints_dict[reaction].append(metabolite)
                else:
                    initial_chokepoints_dict[reaction] = [metabolite]
        #
        chokepoints_dem_dict = {}
        if state_dem.chokepoints() is not None:
            for el in state_dem.chokepoints():
                (reaction, metabolite) = el
                reaction = reaction.id
                metabolite = metabolite.id
                if reaction in chokepoints_dem_dict:
                    chokepoints_dem_dict[reaction].append(metabolite)
                else:
                    chokepoints_dem_dict[reaction] = [metabolite]
        #
        chokepoints_fva_dict = {}
        if state_fva.chokepoints() is not None:
            for el in state_fva.chokepoints():
                (reaction, metabolite) = el
                reaction = reaction.id
                metabolite = metabolite.id
                if reaction in chokepoints_fva_dict:
                    chokepoints_fva_dict[reaction].append(metabolite)
                else:
                    chokepoints_fva_dict[reaction] = [metabolite]
        #
        chokepoints_fva_dem_dict = {}
        if state_fva_dem.chokepoints() is not None:
            for el in state_fva_dem.chokepoints():
                (reaction, metabolite) = el
                reaction = reaction.id
                metabolite = metabolite.id
                if reaction in chokepoints_fva_dem_dict:
                    chokepoints_fva_dem_dict[reaction].append(metabolite)
                else:
                    chokepoints_fva_dem_dict[reaction] = [metabolite]

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = self.__spreadsheet.add_sheet(sheet_name, cell_overwrite_ok=True)
        sheet.write(0, 0, "REACTION ID", style=style)
        sheet.write(0, 1, "REACTION", style=style)
        sheet.write(0, 2, "CHOKEPOINT INITIAL", style=style)
        sheet.write(0, 3, "CHOKEPOINT WITHOUT DEM", style=style)
        sheet.write(0, 4, "CHOKEPOINT AFTER FVA", style=style)
        sheet.write(0, 5, "CHOKEPOINT AFTER FVA AND WITHOUT DEM", style=style)
        i = 1
        for reaction in reactions:
            sheet.write(i, 0, reaction.id)
            sheet.write(i, 1, reaction.reaction)
            j = 0
            if reaction.id in initial_chokepoints_dict:
                for metabolite in initial_chokepoints_dict[reaction.id]:
                    sheet.write(i + j, 2, metabolite)
                    sheet.write(i + j, 6, " ")
                    j = j + 1
            else:
                sheet.write(i + j, 2, " ")
            k = 0
            if reaction.id in chokepoints_dem_dict:
                for metabolite in chokepoints_dem_dict[reaction.id]:
                    sheet.write(i + k, 3, metabolite)
                    sheet.write(i + k, 6, " ")
                    k = k + 1
            else:
                sheet.write(i + k, 3, " ")
            l = 0
            if reaction.id in chokepoints_fva_dict:
                for metabolite in chokepoints_fva_dict[reaction.id]:
                    sheet.write(i + l, 4, metabolite)
                    sheet.write(i + l, 6, " ")
                    l = l + 1
            else:
                sheet.write(i + l, 4, " ")
            m = 0
            if reaction.id in chokepoints_fva_dem_dict:
                for metabolite in chokepoints_fva_dem_dict[reaction.id]:
                    sheet.write(i + m, 5, metabolite)
                    sheet.write(i + m, 6, " ")
                    m = m + 1
            else:
                sheet.write(i + m, 5, " ")
            maxim = max(j, k, l, m)
            if maxim == 0:
                i = i + 1
            else:
                i = i + maxim

    def spreadsheet_write_summary_metabolites(
        self, sheet_name, state_initial, state_fva
    ):

        if self.__spreadsheet is None:
            self.spreadsheet_init()

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font

        aux_dem_init = []
        if state_initial.dem() is not None:
            aux = list(itertools.chain.from_iterable(state_initial.dem().values()))
            aux_dem_init = [e.id for e in aux]
        aux_dem_fva = []
        if state_fva.dem() is not None and state_fva.fva() is not None:
            aux = list(itertools.chain.from_iterable(state_fva.dem().values()))
            aux_dem_fva = [e.id for e in aux]
        metabolites = state_initial.metabolites()
        metabolites.sort(key=self.__id)
        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(0, 0, "METABOLITE ID", style=style)
        sheet.write(0, 1, "METABOLITE NAME", style=style)
        sheet.write(0, 2, "METABOLITE COMPARTMENT", style=style)
        sheet.write(0, 3, "DEM", style=style)
        sheet.write(0, 4, "DEM AFTER FVA", style=style)
        i = 1
        for metabolite in metabolites:
            sheet.write(i, 0, metabolite.id)
            sheet.write(i, 1, metabolite.name)
            sheet.write(i, 2, metabolite.compartment)
            if metabolite.id in aux_dem_init:
                sheet.write(i, 3, "TRUE")
            if metabolite.id in aux_dem_fva:
                sheet.write(i, 4, "TRUE")
            i = i + 1

    def __aux_spreadsheet_write_chokepoints_genes(self, state):
        chokepoints_dict = {}
        if state.chokepoints() is not None:
            for (reaction, metabolite) in state.chokepoints():
                reaction = reaction.id
                metabolite = metabolite.id
                if reaction in chokepoints_dict:
                    chokepoints_dict[reaction].append(metabolite)
                else:
                    chokepoints_dict[reaction] = [metabolite]
        genes_reactions_dict = {}
        if state.essential_genes_reactions() is not None:
            for reaction in state.essential_genes_reactions().keys():
                for gen in state.essential_genes_reactions()[reaction]:
                    reaction = reaction.id
                    gen = gen.id
                    if reaction in genes_reactions_dict:
                        genes_reactions_dict[reaction].append(gen)
                    else:
                        genes_reactions_dict[reaction] = [gen]
        knockout_growth = {}
        if state.knockout_growth() is not None:
            for reaction, growth in state.knockout_growth().items():
                knockout_growth[reaction.id] = growth

        return (chokepoints_dict, genes_reactions_dict, knockout_growth)

    def spreadsheet_write_chokepoints_genes(
        self, sheet_name, state_initial, state_dem, state_fva, state_fva_dem
    ):
        """

        Args:
                sheet_name ():

        Returns:

        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font

        sheet = self.__spreadsheet.add_sheet(sheet_name)
        sheet.write(1, 0, "REACTION ID", style=style)
        sheet.write(0, 2, "INITIAL MODEL", style=style)
        sheet.write(0, 6, "MODEL WITHOUT DEM", style=style)
        sheet.write(0, 10, "MODEL AFTER FVA", style=style)
        sheet.write(0, 14, "MODEL AFTER FVA AND WITHOUT DEM", style=style)

        sheet.write(1, 2, "CHOKEPOINT", style=style)
        sheet.write(1, 3, "ESSENTIAL GENE REACTION", style=style)
        sheet.write(1, 4, "OBJECTIVE VALUE", style=style)
        sheet.write(1, 6, "CHOKEPOINT", style=style)
        sheet.write(1, 7, "ESSENTIAL GENE REACTION", style=style)
        sheet.write(1, 8, "OBJECTIVE VALUE", style=style)
        sheet.write(1, 10, "CHOKEPOINT", style=style)
        sheet.write(1, 11, "ESSENTIAL GENE REACTION", style=style)
        sheet.write(1, 12, "OBJECTIVE VALUE", style=style)
        sheet.write(1, 14, "CHOKEPOINT", style=style)
        sheet.write(1, 15, "ESSENTIAL GENE REACTION", style=style)
        sheet.write(1, 16, "OBJECTIVE VALUE", style=style)

        models = [state_initial, state_dem, state_fva, state_fva_dem]
        j = 2
        reactions = state_initial.reactions()
        reactions.sort(key=self.__id)
        for model in models:
            i = 2
            (
                chokepoints,
                genes_reactions,
                knockout_growth,
            ) = self.__aux_spreadsheet_write_chokepoints_genes(model)
            for reaction_obj in reactions:
                reaction = reaction_obj.id
                if j == 2:
                    sheet.write(i, 0, reaction)
                    sheet.write(i, 1, " ")
                if reaction in chokepoints.keys():
                    sheet.write(i, j, "TRUE")
                if reaction in genes_reactions.keys():
                    sheet.write(i, j + 1, "TRUE")
                if reaction in knockout_growth.keys():
                    sheet.write(i, j + 2, knockout_growth[reaction])
                i = i + 1
            j = j + 4

    def __aux_spreadsheet_get_data(self, state):
        dem = []
        if state.dem() is not None:
            metabolites = list(itertools.chain.from_iterable(state.dem().values()))
            dem = [m.id for m in metabolites]

        chokepoints_dict = {}
        if state.chokepoints() is not None:
            for (reaction, metabolite) in state.chokepoints():
                reaction = reaction.id
                metabolite = metabolite.id
                if reaction in chokepoints_dict:
                    chokepoints_dict[reaction].append(metabolite)
                else:
                    chokepoints_dict[reaction] = [metabolite]
        genes_reactions_dict = {}
        if state.essential_genes_reactions() is not None:
            for reaction in state.essential_genes_reactions().keys():
                for gen in state.essential_genes_reactions()[reaction]:
                    reaction = reaction.id
                    gen = gen.id
                    if reaction in genes_reactions_dict:
                        genes_reactions_dict[reaction].append(gen)
                    else:
                        genes_reactions_dict[reaction] = [gen]
        knockout_growth = {}
        if state.knockout_growth() is not None:
            for reaction, growth in state.knockout_growth().items():
                knockout_growth[reaction.id] = growth

        essential_genes = []
        if state.essential_genes() is not None:
            essential_genes = [g.id for g in state.essential_genes()]

        er_g0 = []
        er_g005 = []
        objective_value = state.objective_value()
        if state.objective_value() is not None:
            if not isnan(state.objective_value()):
                objective_value_005 = state.objective_value() * 0.05
            else:
                objective_value_005 = 0
            for reaction, value in knockout_growth.items():
                if not isnan(objective_value):
                    if isnan(value) or value == 0:
                        er_g0.append(reaction)
                        er_g005.append(reaction)
                    elif value < objective_value_005:
                        er_g005.append(reaction)
                else:
                    er_g0.append(reaction)
                    er_g005.append(reaction)

        reversibles = state.reversible_reactions()

        dead_reactions = state.dead_reactions()

        return [
            dem,
            list(chokepoints_dict.keys()),
            essential_genes,
            genes_reactions_dict,
            knockout_growth,
            er_g0,
            er_g005,
            reversibles,
            dead_reactions,
        ]

    #
    def __aux_spreadsheet_write_two_model_summary(
        self, sheet, style, x, y, title, list_model_1, list_model_2
    ):
        sheet.write(x, y, title, style=style)
        sheet.write(x + 2, y + 1, "Dead-end metabolites", style=style)
        sheet.write(x + 3, y + 1, "Chokepoint reactions", style=style)
        sheet.write(x + 4, y + 1, "Essential genes", style=style)
        sheet.write(x + 5, y + 1, "Essential genes reactions", style=style)
        sheet.write(x + 6, y + 1, "Essential reactions (objective=0)", style=style)
        sheet.write(
            x + 7,
            y + 1,
            "Essential reactions (objective < 5% max objective)",
            style=style,
        )
        sheet.write(x + 8, y + 1, "Reversible reactions", style=style)
        sheet.write(x + 9, y + 1, "Dead reactions", style=style)
        sheet.write(x + 1, y + 2, "Before", style=style)
        sheet.write(x + 1, y + 3, "After", style=style)
        sheet.write(x + 1, y + 5, "Only before", style=style)
        sheet.write(x + 1, y + 6, "Intersection", style=style)
        sheet.write(x + 1, y + 7, "Only after", style=style)

        for i in range(0, len(list_model_1)):
            sheet.write(x + i + 2, y + 2, len(list_model_1[i]))
            sheet.write(x + i + 2, y + 3, len(list_model_2[i]))

        for i in range(0, len(list_model_1)):
            if isinstance(list_model_1[0], list):
                d1 = set(list_model_1[i])
                d2 = set(list_model_2[i])
                set_1 = d1.difference(d2)
                set_2 = d1.intersection(d2)
                set_3 = d2.difference(d1)
                sheet.write(x + i + 2, y + 5, str(len(set_1)))
                sheet.write(x + i + 2, y + 6, str(len(set_2)))
                sheet.write(x + i + 2, y + 7, str(len(set_3)))
            else:
                d1 = list_model_1[i].keys()
                d2 = list_model_2[i].keys()
                set_1 = d1 - d2
                set_2 = d1 & d2
                set_3 = d2 - d1
                sheet.write(x + i + 3, y + 5, str(len(set_1)))
                sheet.write(x + i + 3, y + 6, str(len(set_2)))
                sheet.write(x + i + 3, y + 7, str(len(set_3)))

    def sets_intersection(self, set_list):
        aux = copy(set_list[0])
        for i in range(1, len(set_list)):
            aux = aux.intersection(copy(set_list[i]))
        return aux

    def sets_union(self, set_list):
        if set_list == []:
            return set()
        else:
            aux = set_list[0]
            for i in range(1, len(set_list)):
                aux = aux.union(set_list[i])
            return aux

    def __aux_write_sets(self, sheet, x, y, ls):
        for i in range(len(ls)):
            st = ls.pop(i)
            val = st.difference(self.sets_union(ls))
            sheet.write(x, y + i + 2, len(val))
            ls.insert(i, st)
        val = self.sets_intersection(ls)
        sheet.write(x, y + 6, len(val))
        val = self.sets_union(ls)
        sheet.write(x, y + 7, len(val))

    def __aux_spreadsheet_write_four_model_summary(
        self, sheet, style, x, y, title, model_results
    ):
        sheet.write(x, y, title, style=style)
        sheet.write(x + 2, y + 1, "Dead-end metabolites", style=style)
        sheet.write(x + 3, y + 1, "Chokepoint reactions", style=style)
        sheet.write(x + 4, y + 1, "Essential genes", style=style)
        sheet.write(x + 5, y + 1, "Essential genes reactions", style=style)
        sheet.write(x + 6, y + 1, "Essential reactions (objective=0)", style=style)
        sheet.write(
            x + 7,
            y + 1,
            "Essential reactions (objective < 5% max objective)",
            style=style,
        )
        sheet.write(x + 8, y + 1, "Reversible reactions", style=style)
        sheet.write(x + 9, y + 1, "Dead reactions", style=style)

        sheet.write(x + 1, y + 2, "Only: M0", style=style)
        sheet.write(x + 1, y + 3, "Only: MNODEM", style=style)
        sheet.write(x + 1, y + 4, "Only: MFVA", style=style)
        sheet.write(x + 1, y + 5, "Only: MFVANODEM", style=style)
        sheet.write(x + 1, y + 6, "Intersection", style=style)
        sheet.write(x + 1, y + 7, "Union 4 models", style=style)

        # turn data into sets to operate with
        for model in model_results:
            for i in range(0, len(model)):
                if isinstance(model[i], list):
                    model[i] = set(model[i])
                else:
                    # instance dict
                    model[i] = set(model[i].keys())

        for i in range(0, len(model)):
            set_list = [
                model_results[0][i],
                model_results[1][i],
                model_results[2][i],
                model_results[3][i],
            ]
            if i == 8:
                print(set_list)
            self.__aux_write_sets(sheet, x + i + 2, y, set_list)

    #
    def __aux_write_four_model_two_summary(
        self,
        sheet,
        style,
        x,
        y,
        title,
        subtitle1,
        list_set1,
        subtitle2,
        list_set2,
        subtitle3=None,
        list_set3=None,
    ):

        sheet.write(x, y, title, style=style)
        sheet.write(x + 2, y + 1, subtitle1, style=style)
        sheet.write(x + 3, y + 1, subtitle2, style=style)
        if subtitle3 is None:
            sheet.write(x + 4, y + 1, "Intersection", style=style)
            sheet.write(x + 5, y + 1, "Union", style=style)
        else:
            sheet.write(x + 4, y + 1, subtitle3, style=style)
            sheet.write(x + 5, y + 1, "Intersection", style=style)
            sheet.write(x + 6, y + 1, "Union", style=style)
        sheet.write(x + 1, y + 2, "M0", style=style)
        sheet.write(x + 1, y + 3, "MNODEM", style=style)
        sheet.write(x + 1, y + 4, "MFVA", style=style)
        sheet.write(x + 1, y + 5, "MFVANODEM", style=style)

        i = 0
        for aux_y in range(2, 6):
            if subtitle3 is None:
                if list_set1[i] is not None and list_set2[i] is not None:
                    sheet.write(
                        x + 2, aux_y, len(list_set1[i].difference(list_set2[i]))
                    )
                    sheet.write(
                        x + 3, aux_y, len(list_set2[i].difference(list_set1[i]))
                    )
                    sheet.write(
                        x + 4, aux_y, len(list_set1[i].intersection(list_set2[i]))
                    )
                    sheet.write(x + 5, aux_y, len(list_set1[i].union(list_set2[i])))
                else:
                    for aux_x in range(2, 6):
                        sheet.write(x + aux_x, aux_y, "Nan")
            else:
                if (
                    list_set1[i] is not None
                    and list_set2[i] is not None
                    and list_set3[i] is not None
                ):
                    sheet.write(
                        x + 2,
                        aux_y,
                        len(list_set1[i].difference(list_set2[i].union(list_set3[i]))),
                    )
                    sheet.write(
                        x + 3,
                        aux_y,
                        len(list_set2[i].difference(list_set1[i].union(list_set3[i]))),
                    )
                    sheet.write(
                        x + 4,
                        aux_y,
                        len(list_set3[i].difference(list_set2[i].union(list_set1[i]))),
                    )
                    sheet.write(
                        x + 5,
                        aux_y,
                        len(
                            list_set1[i]
                            .intersection(list_set2[i])
                            .intersection(list_set3[i])
                        ),
                    )
                    sheet.write(
                        x + 6,
                        aux_y,
                        len(list_set1[i].union(list_set2[i]).union(list_set3[i])),
                    )
                else:
                    for aux_x in range(2, 7):
                        sheet.write(x + aux_x, aux_y, "Nan")
            i = i + 1

    def spreadsheet_write_summary(
        self, sheet_name, state_initial, state_dem, state_fva, state_fva_dem
    ):
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font

        sheet = self.__spreadsheet.add_sheet(sheet_name)

        list_model_initial_results = self.__aux_spreadsheet_get_data(state_initial)
        list_model_fva_results = self.__aux_spreadsheet_get_data(state_fva)
        list_model_dem_results = self.__aux_spreadsheet_get_data(state_dem)
        list_model_fvadem_results = self.__aux_spreadsheet_get_data(state_fva_dem)
        models = [
            list_model_initial_results,
            list_model_dem_results,
            list_model_fva_results,
            list_model_fvadem_results,
        ]

        # remove essential_reactions list with objective values
        for m in models:
            del m[4]

        list_metabolites = [
            len(state_initial.metabolites()),
            len(state_dem.metabolites()),
            len(state_fva.metabolites()),
            len(state_fva_dem.metabolites()),
        ]
        list_reactions = [
            len(state_initial.reactions()),
            len(state_dem.reactions()),
            len(state_fva.reactions()),
            len(state_fva_dem.reactions()),
        ]

        # 1: list(chokepoints_dict.keys()),
        # 3: genes_reactions_dict,
        # 5: er_g0
        reactions_data = []
        for i in [1, 3, 5]:
            rdl = [
                list_model_initial_results[i],
                list_model_dem_results[i],
                list_model_fva_results[i],
                list_model_fvadem_results[i],
            ]
            for j in range(len(rdl)):
                if isinstance(rdl[j], list):
                    rdl[j] = set(rdl[j])
                else:
                    rdl[j] = set(rdl[j].keys())
            reactions_data.append(rdl)

        # First, we write model acronyms
        x = 0
        y = 0
        sheet.write(y + 0, x + 0, "ACRONYMS", style=style)
        sheet.write(y + 1, x + 1, "M0", style=style)
        sheet.write(y + 2, x + 1, "MFVA", style=style)
        sheet.write(y + 3, x + 1, "MNODEM", style=style)
        sheet.write(y + 4, x + 1, "MFVANODEM", style=style)

        sheet.write(y + 1, x + 2, "Model in the SBML file")
        sheet.write(y + 2, x + 2, "Model M0 refined with FVA, i.e. with flux bounds constrained with FVA bounds")
        sheet.write(y + 3, x + 2, "Model M0 without dead-end metabolites")
        sheet.write(y + 4, x + 2, "Model M0 refined with FVA and without dead-end metabolites")

        x = 0
        y = y + 6
        sheet.write(y, x, "MODELS DATA", style=style)
        sheet.write(y + 2, x + 1, "Metabolites", style=style)
        sheet.write(y + 3, x + 1, "Reactions", style=style)
        sheet.write(y + 1, x + 2, "M0", style=style)
        sheet.write(y + 1, x + 3, "MNODEM", style=style)
        sheet.write(y + 1, x + 4, "MFVA", style=style)
        sheet.write(y + 1, x + 5, "MFVANODEM", style=style)

        for m, r in zip(list_metabolites, list_reactions):
            sheet.write(y + 2, x + 2, m)
            sheet.write(y + 3, x + 2, r)
            x = x + 1

        x = 0
        y = y + 5 # blank space

        self.__aux_spreadsheet_write_two_model_summary(
            sheet,
            style,
            y,
            0,
            "M0 - MFVA",
            list_model_initial_results,
            list_model_fva_results,
        )

        y = y + 11
        self.__aux_spreadsheet_write_two_model_summary(
            sheet,
            style,
            y,
            0,
            "M0 - MNODEM",
            list_model_initial_results,
            list_model_dem_results,
        )

        y = y + 11
        self.__aux_spreadsheet_write_two_model_summary(
            sheet,
            style,
            y,
            0,
            "MFVA - MFVANODEM",
            list_model_fva_results,
            list_model_fvadem_results,
        )

        y = y + 11
        self.__aux_spreadsheet_write_four_model_summary(
            sheet, style, y, 0, "FOUR MODEL COMPARISON", models
        )

        y = y + 11
        self.__aux_write_four_model_two_summary(
            sheet,
            style,
            y,
            0,
            "CHOKEPOINT REACTIONS - ESSENTIAL GENES REACTIONS",
            "Only chokepoint reaction",
            reactions_data[0],
            "Only essential genes reactions",
            reactions_data[1],
        )

        y = y + 7
        self.__aux_write_four_model_two_summary(
            sheet,
            style,
            y,
            0,
            "CHOKEPOINT REACTIONS - ESSENTIAL REACTIONS (objectite value < 5% max objective value)",
            "Only chokepoints",
            reactions_data[0],
            "Only essential reactions",
            reactions_data[2],
        )

        y = y + 7
        self.__aux_write_four_model_two_summary(
            sheet,
            style,
            y,
            0,
            "ESSENTIAL GENES REACTIONS - ESSENTIAL REACTIONS (objectite value < 5% max objective value)",
            "Only essential genes reactions",
            reactions_data[1],
            "Only essential reactions",
            reactions_data[2],
        )

        y = y + 7
        self.__aux_write_four_model_two_summary(
            sheet,
            style,
            y,
            0,
            "CHOKEPOINT REACTIONS - ESSENTIAL GENES REACTIONS - ESSENTIAL REACTIONS (objectite value < 5% max objective value)",
            "Only chokepoints reactions",
            reactions_data[0],
            "Only essential genes reactions",
            reactions_data[1],
            "Only essential reactions",
            reactions_data[2],
        )


    def spreadsheet_save_file(self, filename):
        """Saves the xlwt Workbook '__spreadsheet' to a valid file.

        Args:
                filename (): output file.
        """
        if self.__spreadsheet is None:
            self.spreadsheet_init()

        if (
            (filename[-4:] == ".xls")
            or (filename[-4:] == ".ods")
            or (filename[-5:] == ".xlsx")
        ):
            self.__spreadsheet.save(filename)
        else:
            raise Exception("File must be: '.xls', '.ods', '.xlsx'")



from __future__ import absolute_import

import os
import click

from contrabass.cli.utils import (
    validate_input_model,
    validate_output_model,
    validate_output_spread,
    save_final_spreadsheet,
    save_final_model,
    check_error_save_final_model,
    verbose_f,
    validate_number,
    validate_output_html_path,
)
from contrabass.core import Facade
from contrabass.core import FacadeUtils
from contrabass.core.utils.GrowthDependentCPConfig import GrowthDependentCPConfig
from contrabass.core.utils.CriticalCPConfig import CriticalCPConfig

TASK_CRITICAL_REACTIONS = "TASK_CRITICAL_REACTIONS"
TASK_SENSIBILITY = "TASK_SENSIBILITY"
TASK_SAVE_WITHOUT_DEM = "TASK_SAVE_WITHOUT_DEM"
TASK_SAVE_WITH_FVA = "TASK_SAVE_WITH_FVA"
TASK_SAVE_WITH_FVA_DEM = "TASK_SAVE_WITH_FVA_DEM"

DEFAULT_SPREADSHEET_FILE = "output.xls"
DEFAULT_HTML_FILE = "output.html"
DEFAULT_MODEL_FILE = "output.xml"

DEFAULT_SPREADSHEET_EXTENSION = ".xls"
HTML_EXTENSION = ".html"

LICENSE = """
    contrabass Copyright (C) 2020-2021  Alex Oarga  <718123 at unizar dot es> (University of Zaragoza)
    This program comes with ABSOLUTELY NO WARRANTY;
    This is free software, and you are welcome to redistribute it
    under certain conditions;
    For details see <https://github.com/openCONTRABASS/CONTRABASS/blob/master/LICENSE>
    If not available, see <https://www.gnu.org/licenses/>
"""


@click.group()
@click.help_option("--help", "-h")
def report():
    """
    Compute vulnerabilities on constrained-based models.
    """
    pass


@click.group()
@click.help_option("--help", "-h")
def new_model():
    """
    Export refined constrained-based model.
    """
    pass


@report.command()
@click.help_option("--help", "-h")
@click.argument(
    "model",
    type=click.Path(exists=True, dir_okay=False),
    # help="Input constrained.based model. Allowed file formats: .xml .json .yml"
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    default=False,
    help="Do not print outputs while running.",
)
@click.option(
    "--objective",
    type=str,
    metavar="OBJECTIVE",
    help="Reaction id to be used as objective function with Flux Balance Analysis",
)
@click.option(
    "--output-spreadsheet",
    "-o",
    type=str,
    metavar="OUTPUT_SPREADSHEET",
    help="Output spreadsheet file with results. Allowed file formats: '.xls' '.xlsx' '.ods'. Default value: "
         "<INPUT-MODEL>.xls",
)
@click.option(
    "--output-html-report",
    "-r",
    type=str,
    metavar="OUTPUT_SPREADSHEET",
    help="Output spreadsheet file with results. Allowed file formats: '.html'. Default value: <INPUT-MODEL>.html",
)
@click.option(
    "--fraction",
    type=float,
    default=1.0,
    metavar="FRACTION",
    help="Fraction of optimum growth to be used in Flux Variability Analysis. Value must be between 0.0 and 1.0",
)
def critical_reactions(
        model, output_spreadsheet, output_html_report, silent, objective, fraction
):
    # Default values
    if fraction is None:
        fraction = 1.0
    fraction = validate_number(fraction)

    if output_spreadsheet is None:
        output_spreadsheet = os.path.splitext(model)[0] + DEFAULT_SPREADSHEET_EXTENSION

    if output_html_report is None:
        output_html_report = os.path.splitext(model)[0] + HTML_EXTENSION

    validate_input_model(model)
    validate_output_spread(output_spreadsheet)
    validate_output_html_path(output_html_report)

    task = TASK_CRITICAL_REACTIONS
    run(
        task,
        model,
        output_spreadsheet,
        output_html_report,
        not silent,
        objective,
        fraction,
    )


@report.command()
@click.help_option("--help", "-h")
@click.argument(
    "model",
    type=click.Path(exists=True, dir_okay=False),
    # help="Input constrained.based model. Allowed file formats: .xml .json .yml"
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    default=False,
    help="Do not print outputs while running.",
)
@click.option(
    "--objective",
    type=str,
    metavar="OBJECTIVE",
    help="Reaction id to be used as objective function with Flux Balance Analysis",
)
@click.option(
    "--output-spreadsheet",
    "-o",
    type=str,
    metavar="OUTPUT_SPREADSHEET",
    help="Output spreadsheet file with results. Allowed file formats: '.xls' '.xlsx' '.ods'. Default value: <INPUT-MODEL>.xls",
)
@click.option(
    "--output-html-report",
    "-r",
    type=str,
    metavar="OUTPUT_HTML",
    help="Output spreadsheet file with results. Allowed file formats: '.html'. Default value: <INPUT-MODEL>.html",
)
def growth_dependent_reactions(
        model,
        output_spreadsheet,
        output_html_report,
        silent,
        objective,
):

    if output_spreadsheet is None:
        output_spreadsheet = os.path.splitext(model)[0] + DEFAULT_SPREADSHEET_EXTENSION

    if output_html_report is None:
        output_html_report = os.path.splitext(model)[0] + HTML_EXTENSION

    validate_input_model(model)
    validate_output_spread(output_spreadsheet)
    validate_output_html_path(output_html_report)

    task = TASK_SENSIBILITY
    run(task, model, output_spreadsheet, output_html_report, not silent, objective, None)


@new_model.command()
@click.help_option("--help", "-h")
@click.argument(
    "model",
    type=click.Path(exists=True, dir_okay=False),
    # help="Input constrained.based model. Allowed file formats: .xml .json .yml"
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    default=False,
    help="Do not print outputs while running.",
)
@click.option(
    "--objective",
    type=str,
    metavar="OBJECTIVE",
    help="Reaction id to be used as objective function with Flux Balance Analysis",
)
@click.option(
    "--fraction",
    type=float,
    default=1.0,
    metavar="FRACTION",
    help="Fraction of optimum growth to be used in Flux Variability Analysis. Value must be between 0.0 and 1.0",
)
@click.option(
    "--output-model",
    "-o",
    type=str,
    metavar="OUTPUT-MODEL",
    help="Output model file. Allowed file formats: '.xml' '.json' '.yml'. Default value: "
         + "<INPUT-MODEL>-MFVA",
)
def fva_constrained(
        model,
        silent,
        objective,
        fraction,
        output_model,
):
    """
    Export a new model with its (maximum and minimum )flux bounds replaced with the
    constrained flux values obtained with Flux Variability Analysis.
    """

    if output_model is None:
        output_model = os.path.splitext(model)[0] + "-MFVA" + ".xml"

    validate_input_model(model)
    validate_output_model(output_model)

    task = TASK_SAVE_WITH_FVA
    run(task, model, output_model, None, not silent, objective, fraction)


@new_model.command()
@click.help_option("--help", "-h")
@click.argument(
    "model",
    type=click.Path(exists=True, dir_okay=False),
    # help="Input constrained.based model. Allowed file formats: .xml .json .yml"
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    default=False,
    help="Do not print outputs while running.",
)
@click.option(
    "--objective",
    type=str,
    metavar="OBJECTIVE",
    help="Reaction id to be used as objective function with Flux Balance Analysis",
)
@click.option(
    "--fraction",
    type=float,
    default=1.0,
    metavar="FRACTION",
    help="Fraction of optimum growth to be used in Flux Variability Analysis. Value must be between 0.0 and 1.0",
)
@click.option(
    "--output-model",
    "-o",
    type=str,
    metavar="OUTPUT-MODEL",
    help="Output model file. Allowed file formats: '.xml' '.json' '.yml'. Default value: "
         + "<INPUT-MODEL>-MNODEM",
)
def without_dem(
        model,
        silent,
        objective,
        fraction,
        output_model,
):
    """
    Export a new model where Dead-End Metabolites have been removed.
    Dead-End Metabolites are removed iteratively, this is, if a reactions remains
    without metabolites it is also removed from the model.
    """

    if output_model is None:
        output_model = os.path.splitext(model)[0] + "-MNODEM" + ".xml"

    validate_input_model(model)
    validate_output_model(output_model)

    task = TASK_SAVE_WITHOUT_DEM
    run(task, model, output_model, None, not silent, objective, fraction)


@new_model.command()
@click.help_option("--help", "-h")
@click.argument(
    "model",
    type=click.Path(exists=True, dir_okay=False),
    # help="Input constrained.based model. Allowed file formats: .xml .json .yml"
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    default=False,
    help="Do not print outputs while running.",
)
@click.option(
    "--objective",
    type=str,
    metavar="OBJECTIVE",
    help="Reaction id to be used as objective function with Flux Balance Analysis",
)
@click.option(
    "--fraction",
    type=float,
    default=1.0,
    metavar="FRACTION",
    help="Fraction of optimum growth to be used in Flux Variability Analysis. Value must be between 0.0 and 1.0",
)
@click.option(
    "--output-model",
    "-o",
    type=str,
    metavar="OUTPUT-MODEL",
    help="Output model file. Allowed file formats: '.xml' '.json' '.yml'. Default value: "
         + "<INPUT-MODEL>-MFVANODEM",
)
def fva_constrained_without_dem(
        model,
        silent,
        objective,
        fraction,
        output_model,
):
    """
    Export a new model with its (maximum and minimum) flux bounds replaced with the
    constrained flux values obtained with Flux Variability Analysis.
    In addition, Dead-End Metabolites are removed iteratively after the flux update.
    """

    if output_model is None:
        output_model = os.path.splitext(model)[0] + "-MFVANODEM" + ".xml"

    validate_input_model(model)
    validate_output_model(output_model)

    task = TASK_SAVE_WITH_FVA_DEM
    run(task, model, output_model, None, not silent, objective, fraction)


@click.command()
@click.help_option("--help", "-h")
def license():
    """
    Print license info.
    """
    click.echo(LICENSE)


def run(
        task,
        model_path,
        output_path,  # output spreadsheet report file or in case of new model, the output model path
        output_path_html,
        verbose,
        objective=None,
        fraction=1.0,
):
    try:
        if task == TASK_CRITICAL_REACTIONS:

            if verbose:
                print("Task: Critical reactions report generation")
            config = CriticalCPConfig()
            config.model_path = model_path
            config.print_f = verbose_f
            config.args1 = verbose
            config.args2 = None
            config.output_path_spreadsheet = output_path
            config.output_path_html = output_path_html
            config.objective = objective
            config.fraction = fraction

            facade = Facade()
            facade.generate_critical_cp_report(config)

        elif task == TASK_SENSIBILITY:

            if verbose:
                print("Task: Growth dependent reactions report generation")
            config = GrowthDependentCPConfig()
            config.model_path = model_path
            config.print_f = verbose_f
            config.args1 = verbose
            config.args2 = None
            config.output_path_spreadsheet = output_path
            config.output_path_html = output_path_html
            config.objective = objective

            facade = Facade()
            facade.generate_growth_dependent_report(config)

        elif task == TASK_SAVE_WITHOUT_DEM:
            if verbose:
                print("Task: save model without dead end metabolites")
            facade = Facade()
            facade.find_and_remove_dem(
                False, output_path, verbose_f, model_path, args1=verbose, args2=None
            )
            save_final_model(facade, output_path, verbose_f)

        elif task == TASK_SAVE_WITH_FVA:
            if verbose:
                print("Task: save model refined with Flux Variability Analysis")
            facade = Facade()
            error = facade.run_fva(
                False,
                output_path,
                verbose_f,
                model_path,
                args1=verbose,
                args2=None,
                objective=objective,
                fraction=fraction,
            )
            check_error_save_final_model(error, facade, output_path, verbose_f)

        elif task == TASK_SAVE_WITH_FVA_DEM:
            print(
                "Task: save model refined with Flux Variability Analysis without dead end metabolites"
            )
            facade = Facade()
            error = facade.run_fva_remove_dem(
                False,
                output_path,
                verbose_f,
                model_path,
                args1=verbose,
                args2=None,
                objective=objective,
                fraction=fraction,
            )
            check_error_save_final_model(error, facade, output_path, verbose_f)

        # Leave final blank line
        print("")

    except Exception as error:
        print("Error: Something went wrong:")
        # Important: the error is not raised but instead it is printed.
        print(str(error))
        raise error



import warnings
from collections import defaultdict

import xlwt
import os
import json

from jinja2 import Environment, FileSystemLoader, Template
from datetime import datetime
from depinfo import get_pkg_info
from dotenv import load_dotenv

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

envars = os.path.dirname(os.path.realpath(__file__)) + "/.env"

from contrabass.core.CobraMetabolicModel import CobraMetabolicModel
from contrabass.core.Spreadsheet import Spreadsheet
from contrabass.core.State import StateEncoder
from contrabass.core.utils.CustomLogger import CustomLogger
from contrabass.core.util import *
from contrabass.core.report_utils import *

# import contrabass.core
# VERSION=contrabass.__version__
VERSION = "0.0.0"


class ErrorGeneratingModel(Exception):
    pass


class FacadeUtils:

    __processes = None

    @property
    def processes(self):
        return self.__processes

    @processes.setter
    def processes(self, processes):
        self.__processes = processes

    def __init__(self, processes=None):
        self.__processes = processes

    def __generate_sheet_sensibility_analysis(
        self, xlwt_worbook, results_growth_dependent, model_state_data
    ):

        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        sheet = xlwt_worbook.add_sheet("Growth dependent reactions")

        model = results_growth_dependent["model"]

        sheet.write(0, 0, "Model", style=style)
        sheet.write(1, 0, "Reactions", style=style)
        sheet.write(2, 0, "Metabolites", style=style)
        sheet.write(3, 0, "Genes", style=style)
        sheet.write(4, 0, "Growth rate (mmol g e−1 h e−1)", style=style)
        sheet.write(5, 0, "Objective", style=style)
        sheet.write(7, 0, "Blocked reactions", style=style)
        sheet.write(8, 0, "Essential reactions", style=style)
        sheet.write(9, 0, "Optimal growth essential reactions", style=style)
        sheet.write(0, 1, model.id())
        sheet.write(1, 1, len(model.reactions()))
        sheet.write(2, 1, len(model.metabolites()))
        sheet.write(3, 1, len(model.genes()))
        sheet.write(4, 1, results_growth_dependent["growth"])
        sheet.write(5, 1, results_growth_dependent["objective"])
        sheet.write(7, 1, len(results_growth_dependent["blocked"]))
        sheet.write(8, 1, len(results_growth_dependent["essential_reactions"][0]))  # only get the initial set of essential reactions
        sheet.write(9, 1, len(results_growth_dependent["optimal_essential"]))

        y = 1
        sheet.write(y + 0, 4, "Reversible Reactions (RR)", style=style)
        sheet.write(y + 1, 4, "Non-reversible Reactions (NR)", style=style)
        sheet.write(y + 2, 4, "Dead Reactions (DR)", style=style)
        sheet.write(y + 3, 4, "Chokepoints (CP)", style=style)
        sheet.write(y + 4, 4, "Essential Reactions (ER)", style=style)

        y = 1
        X_OFFSET = 6
        for i in range(0, len(results_growth_dependent["index"])):
            if i == 0:
                sheet.write(
                    0, i + X_OFFSET, results_growth_dependent["index"][i], style=style
                )
            else:
                sheet.write(
                    0,
                    i + X_OFFSET,
                    "gamma = " + results_growth_dependent["index"][i],
                    style=style,
                )
            if (
                type(results_growth_dependent["reversible"][i]) == list
                or type(results_growth_dependent["reversible"][i]) == set
            ):
                sheet.write(
                    y + 0, i + X_OFFSET, len(results_growth_dependent["reversible"][i])
                )
                sheet.write(
                    y + 1,
                    i + X_OFFSET,
                    len(results_growth_dependent["non_reversible"][i]),
                )
                sheet.write(
                    y + 2, i + X_OFFSET, len(results_growth_dependent["dead"][i])
                )
                sheet.write(
                    y + 3, i + X_OFFSET, len(results_growth_dependent["chokepoints"][i])
                )
                sheet.write(
                    y + 4, i + X_OFFSET, len(results_growth_dependent["essential_reactions"][i])
                )
            else:  # if we are directly given the size
                sheet.write(
                    y + 0, i + X_OFFSET, results_growth_dependent["reversible"][i]
                )
                sheet.write(
                    y + 1, i + X_OFFSET, results_growth_dependent["non_reversible"][i]
                )
                sheet.write(y + 2, i + X_OFFSET, results_growth_dependent["dead"][i])
                sheet.write(
                    y + 3, i + X_OFFSET, results_growth_dependent["chokepoints"][i]
                )
                sheet.write(
                    y + 4, i + X_OFFSET, results_growth_dependent["essential_reactions"][i]
                )

        # generate basic data sheets
        s = Spreadsheet()
        s.set_workbook(xlwt_worbook)
        s.spreadsheet_write_reactions(model_state_data, "reactions")
        s.spreadsheet_write_metabolites(model_state_data, "metabolites")
        s.spreadsheet_write_genes(model_state_data, "genes")
        xlwt_worbook = s.get_workbook()

        return xlwt_worbook

    def __generate_model_info_dict(self, model):
        # Save metadata
        results = {}
        results["solver"] = str(type(model.model().solver))
        results["CONTRABASS_version"] = VERSION
        results["date"] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        # save model data
        results["model_identifier"] = model.id()
        results["growth"] = model.get_growth()
        results["objective"] = model.objective()
        return results

    def compute_growth_dependent_chokepoints(
        self, model_path, print_f, arg1, arg2, objective=None
    ):

        FRAC = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.99, 1.0]
        results_growth_dependent = defaultdict(list)
        logger = CustomLogger(print_f, arg1, arg2)

        logger.print("Reading model...")
        model = read_model(model_path, objective=objective, processes=self.__processes)
        model.save_state("initial") # we save the initial state just to get basic informatoin from the model later
        initial_state_data = model.get_state("initial")

        logger.print("Computing essential reactions...")
        model.knockout_reactions_growth()
        model.compute_essential_reactions()
        logger.print("Computing optimal growth essential reactions")
        model.find_optimal_growth_essential_reactions()

        reversible_initial = set([r.id for r in model.reversible_reactions()])
        dead_reactions_initial = set([r.id for r in model.dead_reactions()])
        non_reversible_initial = (
            set([r.id for r in model.reactions()])
            .difference(dead_reactions_initial)
            .difference(reversible_initial)
        )
        model.find_chokepoints(exclude_dead_reactions=True)
        chokepoints_initial = set(model.chokepoint_reactions())

        essential_reactions_initial = set([r.id for r in model.essential_reactions()])

        # save model data
        results_growth_dependent["model"] = model
        results_growth_dependent["reactions"] = simplified_reactions(model.reactions())
        results_growth_dependent["metabolites"] = simplified_metabolites(model.metabolites())
        results_growth_dependent["genes"] = [g.id for g in model.genes()]
        results_growth_dependent["compartments"] = list(model.compartments())
        results_growth_dependent["essential"] = [r.id for r in model.knockout_growth()]
        results_growth_dependent["optimal_essential"] = [
            r.id for r in model.optimal_growth_essential_reactions()
        ]
        # growth dependent values
        results_growth_dependent["index"].append("Initial")
        results_growth_dependent["reversible"].append(list(reversible_initial))
        results_growth_dependent["dead"].append(list(dead_reactions_initial))
        results_growth_dependent["non_reversible"].append(list(non_reversible_initial))
        results_growth_dependent["chokepoints"].append(list(chokepoints_initial))
        results_growth_dependent["essential_reactions"].append(
            list(essential_reactions_initial)
        )
        results_growth_dependent["fva_result"].append([])

        # growth essential reactions are computed on the initial model, as FVA is not mandatory for the computation
        # and one execution of the knock-out of each reaction (method) is enough.
        # To summarise: growth dependent essential reactions are computed comparing against a fraction of optimal growth
        logger.print("Computing growth dependent essential reactions")
        for i in range(0, len(FRAC)):
            model.find_growth_essential_reactions(FRAC[i])
            growth_essential_reactions = set(
                [r.id for r in model.growth_essential_reactions()]
            )
            results_growth_dependent["essential_reactions"].append(
                list(growth_essential_reactions)
            )

        y = 1
        for i in range(0, len(FRAC)):

            model = read_model(
                model_path, objective=objective, processes=self.__processes
            )

            logger.print(
                "Running Flux Variability Analysis with fraction: " + str(FRAC[i])
            )
            errors_fva = model.fva(update_flux=True, threshold=FRAC[i])

            if errors_fva != []:
                logger.print(
                    "Could not run Flux Variability Analysis: " + str(errors_fva[0])
                )
                err_msg = "Error running FVA: " + str(errors_fva[0])
                results_growth_dependent["error"] = err_msg
                results_growth_dependent["index"].append(str(FRAC[i]))
                results_growth_dependent["reversible"].append(err_msg)
                results_growth_dependent["dead"].append(err_msg)
                results_growth_dependent["non_reversible"].append(err_msg)
                results_growth_dependent["chokepoints"].append(err_msg)
                results_growth_dependent["fva_result"].append(err_msg)
                continue

            fva_dead_reactions = set([r.id for r in model.dead_reactions()])
            fva_reversible = set([r.id for r in model.reversible_reactions()])
            non_reversible = (
                set([r.id for r in model.reactions()])
                .difference(fva_reversible)
                .difference(fva_dead_reactions)
            )
            model.find_chokepoints(exclude_dead_reactions=True)
            chokepoints = set(model.chokepoint_reactions())

            results_growth_dependent["index"].append(str(FRAC[i]))
            results_growth_dependent["reversible"].append(list(fva_reversible))
            results_growth_dependent["dead"].append(list(fva_dead_reactions))
            results_growth_dependent["non_reversible"].append(list(non_reversible))
            results_growth_dependent["chokepoints"].append(list(chokepoints))
            results_growth_dependent["fva_result"].append(
                [(r.id, upper, lower) for (r, upper, lower) in model.get_fva()]
            )

        # Blocked reactions are equal to dead reactions with gamma = 0.0 (i.e. fraction of optimum = 0.0)
        results_growth_dependent["blocked"] = results_growth_dependent["dead"][1]

        # generate generic model data
        results_growth_dependent = {
            **results_growth_dependent,
            **self.__generate_model_info_dict(model),
        }

        return results_growth_dependent, initial_state_data

    def generate_growth_dependent_spreadsheet(self, growth_dependent_results, state):
        s = xlwt.Workbook()
        s = self.__generate_sheet_sensibility_analysis(s, growth_dependent_results, state)
        sObject = Spreadsheet()
        sObject.set_workbook(s)

        return sObject

    def generate_growth_dependent_html_report(self, results_growth_dependent):
        # cobra model not json serializable
        del results_growth_dependent["model"]
        return generate_html_template(results_growth_dependent, "template.html")

    def generate_critical_cp_html_report(self, critical_cp_results):
        data = {}
        encoder = StateEncoder()
        data["initial"] = encoder.encode(critical_cp_results["initial"])
        data["fva"] = encoder.encode(critical_cp_results["fva"])
        data["dem"] = encoder.encode(critical_cp_results["dem"])
        data["fva_dem"] = encoder.encode(critical_cp_results["fva_dem"])
        return generate_html_template(
            critical_cp_results, "template_cp.html", default=encoder.default
        )

    def run_sensibility_analysis(self, model_path, print_f, arg1, arg2, objective=None):
        warnings.warn(
            "run_sensibility_analysis() is deprecated, use instead:\n"
            + "    results = facadeUtils.compute_growth_dependent_chokepoints(...) \n"
            + "    xlwt_workbook = facadeUtils.generate_growth_dependent_spreadsheet(result)",
            DeprecationWarning,
        )
        results_growth_dependent, model_state_data = self.compute_growth_dependent_chokepoints(
            model_path, print_f, arg1, arg2, objective
        )

        s = xlwt.Workbook()
        s = self.__generate_sheet_sensibility_analysis(s, results_growth_dependent, model_state_data)
        sObject = Spreadsheet()
        sObject.set_workbook(s)

        return sObject

    def compute_critical_points(
        self, model_path, print_f, arg1, arg2, objective=None, fraction=1.0
    ):
        print_f("Reading model...", arg1, arg2)
        model = CobraMetabolicModel(model_path)
        if objective is not None:
            model.set_objective(objective)
        if self.__processes is not None:
            model.processes = self.__processes

        model.save_state("initial")
        model.save_state("dem")
        model.save_state("fva")
        model.save_state("fva_dem")

        # save general model info
        general_info = self.__generate_model_info_dict(model)

        print_f("Generating models...", arg1, arg2)

        model.find_essential_genes_reactions()
        print_f("Searching Dead End Metabolites (D.E.M.)...", arg1, arg2)
        model.find_dem()
        print_f("Searching chokepoint reactions...", arg1, arg2)
        model.find_chokepoints(exclude_dead_reactions=True)
        print_f("Searching essential reactions...", arg1, arg2)
        model.knockout_reactions_growth()
        print_f("Searching essential genes...", arg1, arg2)
        errors_initial = model.find_essential_genes_1()

        error = None
        if errors_initial != []:
            MSG = "Couldn't find essential genes: " + str(errors_initial[0])
            error = MSG
            print_f(MSG)
        else:
            print_f("Searching essential genes reactions...", arg1, arg2)
            model.find_essential_genes_reactions()

        model.save_state("initial", error=error)

        print_f("Removing Dead End Metabolites (D.E.M.)...", arg1, arg2)
        model.remove_dem()
        print_f("Searching essential reactions...", arg1, arg2)
        model.knockout_reactions_growth()
        print_f("Searching new chokepoint reactions...", arg1, arg2)
        model.find_chokepoints(exclude_dead_reactions=True)

        error = None
        if errors_initial == []:
            print_f("Searching essential genes...", arg1, arg2)
            errors_dem = model.find_essential_genes_1()
            if errors_dem == []:
                print_f("Searching essential genes reactions...", arg1, arg2)
                model.find_essential_genes_reactions()
            else:
                error = errors_dem[0]
        else:
            error = errors_initial[0]

        model.save_state("dem", error=error)

        print_f("Running Flux Variability Analysis...", arg1, arg2)
        model = CobraMetabolicModel(model_path)

        if objective is not None:
            model.set_objective(objective)
        if self.__processes is not None:
            model.processes = self.__processes

        errors_fva = model.fva(update_flux=True, threshold=fraction)

        error = None
        if errors_fva != []:
            MSG = "Could not run Flux Variability Analysis: " + str(errors_fva[0])
            error = MSG
            print_f(MSG, arg1, arg2)
        else:
            print_f("Searching Dead End Metabolites (D.E.M.)...", arg1, arg2)
            model.find_dem()
            print_f("Searching new chokepoint reactions...", arg1, arg2)
            model.find_chokepoints(exclude_dead_reactions=True)
            print_f("Searching essential genes...", arg1, arg2)
            errors_fva_genes = model.find_essential_genes_1()
            if errors_fva_genes != []:
                MSG = "Could not find essential genes: " + str(errors_fva_genes[0])
                error = MSG
                print_f(MSG)
            else:
                print_f("Searching essential genes reactions...", arg1, arg2)
                model.find_essential_genes_reactions()
            print_f("Searching essential reactions...", arg1, arg2)
            model.knockout_reactions_growth()

            model.save_state("fva", error=error)

            print_f("Removing Dead End Metabolites (D.E.M.)...", arg1, arg2)
            model.remove_dem()
            print_f("Searching essential reactions...", arg1, arg2)
            model.knockout_reactions_growth()
            print_f("Searching new chokepoint reactions...", arg1, arg2)
            model.find_chokepoints(exclude_dead_reactions=True)
            error = None
            if errors_fva_genes == []:
                print_f("Searching essential genes...", arg1, arg2)
                model.find_essential_genes_1()
                print_f("Searching essential genes reactions...", arg1, arg2)
                model.find_essential_genes_reactions()
            else:
                error = None

            model.save_state("fva_dem", error=error)

        result = {}
        result["initial"] = model.get_state("initial")
        result["dem"] = model.get_state("dem")
        result["fva"] = model.get_state("fva")
        result["fva_dem"] = model.get_state("fva_dem")

        errors_list_of_each_state = [result["initial"].error(), result["dem"].error(), result["fva"].error(), result["fva_dem"].error()]
        result["error"] = next((el for el in errors_list_of_each_state if el is not None and len(el) > 0), "")

        # include general model info computed previously
        result = {**result, **general_info}

        return result

    def generate_critical_cp_spreadsheet(self, critical_cp_results):

        initial = critical_cp_results["initial"]
        fva = critical_cp_results["fva"]
        dem = critical_cp_results["dem"]
        fva_dem = critical_cp_results["fva_dem"]

        s = Spreadsheet()
        s.spreadsheet_write_model_info(initial, "model_info")
        s.spreadsheet_write_summary(
            "summary",
            initial,
            dem,
            fva,
            fva_dem,
        )
        s.spreadsheet_write_reactions(initial, "reactions", ordered=True)
        s.spreadsheet_write_metabolites(
            initial,
            "metabolites",
            ordered=True,
            print_reactions=True,
        )
        s.spreadsheet_write_genes(initial, "genes", ordered=True, print_reactions=True)

        s.spreadsheet_write_reactions(fva, "reactions_FVA", ordered=True)
        s.spreadsheet_write_metabolites(
            fva,
            "metabolites_FVA",
            ordered=True,
            print_reactions=True,
        )

        s.spreadsheet_write_reversible_reactions(
            "reversible reactions",
            initial,
            fva,
            ordered=True,
        )
        s.spreadsheet_write_summary_reactions(
            "chokepoints",
            initial,
            dem,
            fva,
            fva_dem,
        )
        s.spreadsheet_write_summary_metabolites("dead-end", initial, fva)
        s.spreadsheet_write_chokepoints_genes(
            "comparison",
            initial,
            dem,
            fva,
            fva_dem,
        )
        s.spreadsheet_write_essential_genes_comparison(
            "essential genes",
            initial,
            dem,
            fva,
            fva_dem,
            ordered=True,
        )
        s.spreadsheet_write_essential_reactions(
            "essential reactions",
            initial,
            dem,
            fva,
            fva_dem,
            ordered=True,
        )

        return s

    def run_summary_model(
        self, model_path, print_f, arg1, arg2, objective=None, fraction=1.0
    ):
        warnings.warn(
            "run_summary_model() is deprecated, use instead:\n"
            + "    results = facadeUtils.compute_critical_points(...) \n"
            + "    xlwt_workbook = facadeUtils.generate_critical_cp_spreadsheet(result)",
            DeprecationWarning,
        )

        results = self.compute_critical_points(
            model_path, print_f, arg1, arg2, objective, fraction
        )
        return self.generate_critical_cp_spreadsheet(results)

    def find_and_remove_dem(self, model_path):
        model = CobraMetabolicModel(model_path)
        model.find_dem()
        model.remove_dem()
        return model

    def run_fva(self, model_path, objective=None, fraction=1.0):

        model = CobraMetabolicModel(model_path)

        if objective is not None:
            model.set_objective(objective)
        if self.__processes is not None:
            model.processes = self.__processes

        errors_fva = model.fva(update_flux=True)
        errors = model.fva(update_flux=True, threshold=fraction)
        if errors != []:
            return (model, errors[0])
        else:
            return (model, "")

    def run_fva_remove_dem(self, model_path, objective=None, fraction=1.0):

        model = CobraMetabolicModel(model_path)

        if objective is not None:
            model.set_objective(objective)
        if self.__processes is not None:
            model.processes = self.__processes

        errors = model.fva(update_flux=True, threshold=fraction)
        if errors != []:
            return (model, errors[0])
        else:
            model.find_dem()
            model.remove_dem()
            return (model, "")

    def save_model(self, output_path, model):
        if output_path != "":
            try:
                model.save_model(output_path)
                return (True, output_path)
            except Exception as error:
                if os.environ.get(ENV_ENVIRONMENT) == ENV_DEV:
                    raise error
                return (False, str(error))
        return (False, "")

    def save_spreadsheet(self, output_path, spreadsheet):
        if output_path != "":
            try:
                spreadsheet.spreadsheet_save_file(output_path)
                return (True, output_path)
            except Exception as error:
                if os.environ.get(ENV_ENVIRONMENT) == ENV_DEV:
                    raise error
                return (False, str(error))
        return (False, "")

    def read_model(self, model_path):
        try:
            model = CobraMetabolicModel(model_path)
            model_id = model.id()
            reactions = len(model.reactions())
            metabolites = len(model.metabolites())
            genes = len(model.genes())
            reactions_list = [x.id for x in model.reactions()]
            return (
                True,
                None,
                model,
                model_id,
                reactions,
                metabolites,
                genes,
                reactions_list,
            )
        except Exception as error:
            print(error)
            return (False, str(error), None, None, None, None, None, None)

    def print_something(self):
        pass



import os, json
from jinja2 import Environment, FileSystemLoader, Template

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from contrabass.core.templates import html


def format_annotation(cobra_object):
    return [k + " : " + str(v) for (k,v) in cobra_object.annotation.items()]



def simplified_reactions(reaction_list):
    """
    Args:
        reaction_list: List of cobra reactions
    Returns: Turns reactions list into list of dictionaries with each dict containing fields: id, name, formula, annotations
    """
    return [{
        "id": r.id,
        "name": r.name,
        "formula": "" if r.reaction is None else r.reaction,
        "annotations": format_annotation(r)
    } for r in reaction_list]


def simplified_metabolites(metabolites_list):
    """
    Args:
        metabolites_list: List of cobra metabolites
    Returns: Turns metabolites list into list of dictionaries with each dict containing fields: id, name, formula, annotations
    """
    return [{
        "id": r.id,
        "name": r.name,
        "formula": "" if r.formula is None else r.formula,
        "annotations": format_annotation(r)
    } for r in metabolites_list]


def simplified_genes(genes_list):
    """
    Args:
        metabolites_list: List of cobra genes
    Returns: Turns genes list into list of dictionaries with each dict containing fields: id, name, annotations
    """
    return [{
        "id": r.id,
        "name": r.name,
        "annotations": format_annotation(r)
    } for r in genes_list]


def generate_html_template(data, template_name, default=None):
    # file_loader = FileSystemLoader(searchpath="./")
    # env = Environment(loader=file_loader)
    # template = env.get_template('template.html')

    html_template = pkg_resources.read_text(html, template_name)

    """
    jinja2 is giving error:
        jinja2.exceptions.TemplateSyntaxError: unexpected char '\\' at 280624
    on angular-produced template in the script section
    The following quickfix is implemented as solution: 
        scripts are replaced and then restored after template is parsed.
    """
    START_SCRIPT_HTML = "<script"
    END_SCRIPT_HTML = "</script>"
    script_sections = []
    for i in range(0, 3):
        script_start = html_template.find(START_SCRIPT_HTML)
        script_end = html_template.find(END_SCRIPT_HTML)
        script_section = html_template[script_start : script_end + len(END_SCRIPT_HTML)]
        script_sections.append(script_section)
        no_script_template = (
            html_template[:script_start]
            + "{{ script_"
            + str(i)
            + " }}"
            + html_template[script_end + len(END_SCRIPT_HTML) :]
        )
        html_template = no_script_template

    # include vars placeholder
    script_start = no_script_template.find("{{ script_0 }}")
    no_script_template = (
        no_script_template[:script_start]
        + "<script>window.data = {{ results }};</script>"
        + no_script_template[script_start:]
    )

    # Uncomment me to get test data json string
    #print(json.dumps(data, default=default).replace("NaN", "\"NaN\""))

    template = Template(no_script_template)
    output = template.render(
        results=json.dumps(data, default=default).replace("NaN", '"NaN"'),
        script_0=script_sections[0],
        script_1=script_sections[1],
        script_2=script_sections[2],
    )

    return output



# imports on release
from contrabass.core.CobraMetabolicModel import CobraMetabolicModel



def read_model(model_path, objective=None, processes=None):
    model = CobraMetabolicModel(model_path)
    if objective is not None:
        model.set_objective(objective)
    if processes is not None:
        model.processes = processes
    return model


def write_file(path, string):
    text_file = open(path, "w")
    n = text_file.write(string)
    text_file.close()


def read_file(path):
    with open(path, "r") as file:
        data = file.read()
        return data

